from functools import partial
from typing import Union, Dict, TYPE_CHECKING

from mewpy.util.constants import ModelConstants

from mewpy.germ.analysis import FBA
from mewpy.germ.lp import ConstraintContainer, VariableContainer, concat_constraints, integer_coefficients
from mewpy.germ.solution import ModelSolution
from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel
from mewpy.solvers import Solution
from mewpy.solvers.solver import Solver, VarType

if TYPE_CHECKING:
    from mewpy.germ.variables import Reaction, Interaction


class SRFBA(FBA):

    def __init__(self,
                 model: Union[Model, MetabolicModel, RegulatoryModel],
                 solver: Union[str, Solver, None] = None,
                 build: bool = False,
                 attach: bool = False):
        """
        Steady-state Regulatory Flux Balance Analysis (SRFBA) of a metabolic-regulatory model.
        Implementation of a steady-state version of SRFBA for an integrated metabolic-regulatory model.

        For more details consult Shlomi et al. 2007 at https://dx.doi.org/10.1038%2Fmsb4100141

        :param model: a MetabolicModel, RegulatoryModel or GERM model. The model is used to retrieve
        variables and constraints to the linear problem
        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used, but it will be overwritten
        if build is true.
        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """
        super().__init__(model=model, solver=solver, build=build, attach=attach)
        self._model_default_lb = ModelConstants.REACTION_LOWER_BOUND
        self._model_default_ub = ModelConstants.REACTION_UPPER_BOUND

    @property
    def model_default_lb(self) -> float:
        """
        The default lower bound for the model reactions.
        :return:
        """
        if self.synchronized:
            return self._model_default_lb

        self._model_default_lb = min(reaction.lower_bound for reaction in self.model.yield_reactions())
        return self._model_default_lb

    @property
    def model_default_ub(self) -> float:
        """
        The default upper bound for the model reactions.
        :return:
        """
        if self.synchronized:
            return self._model_default_ub

        self._model_default_ub = max(reaction.upper_bound for reaction in self.model.yield_reactions())
        return self._model_default_ub

    def gpr_constraint(self, reaction: 'Reaction'):
        """
        It creates a constraint for a given GPR where variables are genes and constraints are designed for
        each reaction.
        A linear GPR follows a boolean algebra-based model.

        The GPR is translated as a set of variables and constraints using the method `linearize_expression`.

        First, it creates the relation between reaction boolean value and the reaction constrains
        For that, the following reactions are added
        V - Y*Vmax <= 0
        V - Y*Vmin => 0
        where V is the reaction in S matrix
        where Y is the reaction boolean variable
        It adds reaction boolean variable and the constraint

        Then,
        Constraints are created using the multiple methods for each operator. Consult `and_constraint`, `or_constraint`,
        `not_constraint`, `greater_constraint`, `less_constraint`, `equal_constraint`, `true_constraint`,
        `false_constraint`, `symbol_constraint` for more information.

        Variables are created using the multiple methods for each operator. Consult `and_variable`, `or_variable`,
        `not_variable`, `greater_variable`, `less_variable`, `equal_variable`, `true_variable`,
        `false_variable`, `symbol_variable` for more information.

        :param reaction: the reaction
        :return: gpr variables and constraints
        """
        boolean_variable = f'bool_{reaction.id}'

        variables = [VariableContainer(name=boolean_variable,
                                       sub_variables=[boolean_variable],
                                       lbs=[0.0],
                                       ubs=[1.0],
                                       variables_type=[VarType.INTEGER])]

        lb, ub = reaction.bounds

        coefs = [{reaction.id: 1.0, boolean_variable: -float(ub)},
                 {reaction.id: 1.0, boolean_variable: -float(lb)}]
        lbs = [self.model_default_lb - float(ub), 0.0]
        ubs = [0.0, self.model_default_ub - float(lb)]

        constraints = [ConstraintContainer(name=None,
                                           coefs=coefs,
                                           lbs=lbs,
                                           ubs=ubs)]

        expression_variables, expression_cnt = self.linearize_expression(boolean_variable=boolean_variable,
                                                                         symbolic=reaction.gpr.symbolic)

        variables.extend(expression_variables)
        constraints.extend(expression_cnt)

        constraints = [concat_constraints(constraints=constraints, name=reaction.id)]
        return variables, constraints

    def interaction_constraint(self, interaction: 'Interaction'):
        """
        It creates a constraint for a given interaction where variables are regulators and constraints are designed for
        each target.
        A linear interaction follows a boolean algebra-based model.

        The interaction is translated as a set of variables and constraints using the method `linearize_expression`.

        Constraints are created using the multiple methods for each operator. Consult `and_constraint`, `or_constraint`,
        `not_constraint`, `greater_constraint`, `less_constraint`, `equal_constraint`, `true_constraint`,
        `false_constraint`, `symbol_constraint` for more information.

        Variables are created using the multiple methods for each operator. Consult `and_variable`, `or_variable`,
        `not_variable`, `greater_variable`, `less_variable`, `equal_variable`, `true_variable`,
        `false_variable`, `symbol_variable` for more information.

        :param interaction: the interaction
        :return: interaction variables and constraints
        """
        symbolic = None
        for coefficient, expression in interaction.regulatory_events.items():

            if coefficient > 0.0:
                symbolic = expression.symbolic

        if symbolic is None:
            return [], []

        lb = float(min(interaction.target.coefficients))
        ub = float(max(interaction.target.coefficients))

        if (lb, ub) in integer_coefficients:
            v_type = VarType.INTEGER
        else:
            v_type = VarType.CONTINUOUS

        variables = [VariableContainer(name=interaction.target.id,
                                       sub_variables=[interaction.target.id],
                                       lbs=[lb],
                                       ubs=[ub],
                                       variables_type=[v_type])]
        constraints = []

        expression_variables, expression_cnt = self.linearize_expression(boolean_variable=interaction.target.id,
                                                                         symbolic=symbolic)

        variables.extend(expression_variables)
        constraints.extend(expression_cnt)

        constraints = [concat_constraints(constraints=constraints, name=interaction.target.id)]
        return variables, constraints

    @staticmethod
    def and_constraint(symbolic):
        """
        Following Boolean algebra, an And (a = b and c) can be translated as: a = b*c
        Alternatively, an And can be written as lb < b + c - a < ub

        So, for the midterm expression a = b and c
        We have therefore the equation: -1 <= 2*b + 2*c – 4*a <= 3
        :param symbolic: the symbolic expression
        :return: a constraint
        """
        name = symbolic.key()
        names = [f'{name}_{i}' for i, _ in enumerate(symbolic.variables[:-1])]

        and_op = names[0]
        op_l = symbolic.variables[0]
        op_r = symbolic.variables[1]

        _coefs = []
        _lbs = []
        _ubs = []

        if op_l.is_one or op_l.is_true:

            _coef = {and_op: -4.0, op_r.key(): 2.0}
            _state = (-3.0, 1.0)

        elif op_r.is_one or op_r.is_true:

            _coef = {and_op: -4.0, op_l.key(): 2.0}
            _state = (-3.0, 1.0)

        elif op_l.is_zero or op_l.is_false:

            _coef = {and_op: -4.0, op_r.key(): 2.0}
            _state = (-1.0, 3.0)

        elif op_r.is_zero or op_r.is_false:

            _coef = {and_op: -4.0, op_l.key(): 2.0}
            _state = (-1.0, 3.0)

        else:

            _coef = {and_op: -4.0, op_l.key(): 2.0, op_r.key(): 2.0}
            _state = (-1.0, 3.0)

        _coefs.append(_coef)
        _lbs.append(_state[0])
        _ubs.append(_state[1])

        children = []

        if len(symbolic.variables) > 2:
            children = symbolic.variables[2:]

        # building a nested And subexpression
        for i, op_r in enumerate(children):

            op_l = names[i]
            and_op = names[i + 1]

            if op_r.is_one or op_r.is_true:

                _coef = {and_op: -4.0, op_l: 2.0}
                _state = (-3.0, 1.0)

            elif op_r.is_zero or op_r.is_false:

                _coef = {and_op: -4.0, op_l: 2.0}
                _state = (-1.0, 3.0)

            else:
                _coef = {and_op: -4.0, op_l: 2.0, op_r.key(): 2.0}
                _state = (-1.0, 3.0)

            _coefs.append(_coef)
            _lbs.append(_state[0])
            _ubs.append(_state[1])

        return ConstraintContainer(name=None, coefs=_coefs, lbs=_lbs, ubs=_ubs)

    @staticmethod
    def or_constraint(symbolic):
        """
        Following Boolean algebra, an Or (a = b or c) can be translated as: a = b + c - b*c
        Alternatively, an Or can be written as lb < b + c - a < ub

        So, for the midterm expression a = b or c
        We have therefore the equation: -2 <= 2*b + 2*c – 4*a <= 1
        :param symbolic: the symbolic expression
        :return: a constraint
        """
        name = symbolic.key()
        names = [f'{name}_{i}' for i, _ in enumerate(symbolic.variables[:-1])]

        or_op = names[0]
        op_l = symbolic.variables[0]
        op_r = symbolic.variables[1]

        _coefs = []
        _lbs = []
        _ubs = []

        if op_l.is_one or op_l.is_true:

            _coef = {or_op: -4.0, op_r.key(): 2.0}
            _state = (-4.0, -1.0)

        elif op_r.is_one or op_r.is_true:

            _coef = {or_op: -4.0, op_l.key(): 2.0}
            _state = (-4.0, -1.0)

        elif op_l.is_zero or op_l.is_false:

            _coef = {or_op: -4.0, op_r.key(): 2.0}
            _state = (-2.0, 0.0)

        elif op_r.is_zero or op_r.is_false:

            _coef = {or_op: -4.0, op_l.key(): 2.0}
            _state = (-2.0, 0.0)

        else:

            _coef = {or_op: -4.0, op_l.key(): 2.0, op_r.key(): 2.0}
            _state = (-2.0, 1.0)

        _coefs.append(_coef)
        _lbs.append(_state[0])
        _ubs.append(_state[1])

        children = []

        if len(symbolic.variables) > 2:
            children = symbolic.variables[2:]

        # building a nested Or subexpression
        for i, op_r in enumerate(children):

            op_l = names[i]
            or_op = names[i + 1]

            if op_r.is_one or op_r.is_true:

                _coef = {or_op: -4.0, op_l: 2.0}
                _state = (-4.0, -1.0)

            elif op_r.is_zero or op_r.is_false:

                _coef = {or_op: -4.0, op_l: 2.0}
                _state = (-2.0, 1.0)

            else:

                _coef = {or_op: -4.0, op_l: 2.0, op_r.key(): 2.0}
                _state = (-2.0, 1.0)

            _coefs.append(_coef)
            _lbs.append(_state[0])
            _ubs.append(_state[1])

        return ConstraintContainer(name=None, coefs=_coefs, lbs=_lbs, ubs=_ubs)

    @staticmethod
    def not_constraint(symbolic):
        """
        Following Boolean algebra, a Not (a = not b) can be translated as: a = 1 - b
        Alternatively, a Not can be written as a + b = 1

        So, for the midterm expression a = not b
        We have therefore the equation: 1 < a + b < 1
        :param symbolic: the symbolic expression
        :return: a constraint
        """
        op_l = symbolic.variables[0]

        # Not right operators
        if op_l.is_numeric:

            _coef = {symbolic.key(): 1.0}
            _state = (float(op_l.value), float(op_l.value))

        else:

            # add Not row and set mip bounds to 1;1
            _coef = {symbolic.key(): 1.0, op_l.key(): 1.0}
            _state = (1.0, 1.0)

        return ConstraintContainer(name=None, coefs=[_coef], lbs=[_state[0]], ubs=[_state[1]])

    def greater_constraint(self, symbolic):
        """
        Following Propositional logic, a predicate (a => r > value) can be translated as: r - value > 0
        Alternatively, flux predicate a => r>value can be a(value + tolerance - r_UB) + r < value + tolerance
        & a(r_LB - value - tolerance) + r > r_LB

        So, for the midterm expression a => r > value
        We have therefore the equations: a(value + tolerance - r_UB) + r < value + tolerance
        & a(r_LB - value - tolerance) + r > r_LB
        :param symbolic: the symbolic expression
        :return: a constraint
        """

        greater_op = symbolic.key()
        op_l = symbolic.variables[0]
        op_r = symbolic.variables[1]

        if op_l.is_numeric:
            operand = op_r
            c_val = float(op_l.value)

        else:
            operand = op_l
            c_val = float(op_r.value)

        _lb, _ub = operand.bounds

        _lb = float(_lb)
        _ub = float(_ub)
        # add Greater row (a(value + tolerance - r_UB) + r < value + tolerance) and set mip bounds to
        # -inf;comparison_val
        # add Greater row (a(r_LB - value - tolerance) + r > r_LB) and set mip bounds to lb;inf
        _coefs = [
            {greater_op: c_val + ModelConstants.TOLERANCE - _ub, operand.key(): 1.0},
            {greater_op: _lb - c_val - ModelConstants.TOLERANCE, operand.key(): 1.0}
        ]

        _lbs = [self.model_default_lb, _lb]
        _ubs = [c_val + ModelConstants.TOLERANCE, self.model_default_ub]

        return ConstraintContainer(name=None, coefs=_coefs, lbs=_lbs, ubs=_ubs)

    def less_constraint(self, symbolic):
        """
        Following Propositional logic, a| predicate (a => r < value) can be translated as: r - value < 0
        Alternatively, flux predicate a => r<value can be a(value + tolerance - r_LB) + r > value + tolerance
        & a(r_UB - value - tolerance) + r < r_UB

        So, for the midterm expression a => r > value
        We have therefore the equations: a(value + tolerance - r_LB) + r > value + tolerance
        & a(r_UB - value - tolerance) + r < r_UB
        :param symbolic: the symbolic expression
        :return: a constraint
        """
        less_op = symbolic.key()
        op_l = symbolic.variables[0]
        op_r = symbolic.variables[1]

        if op_l.is_numeric:
            operand = op_r
            c_val = float(op_l.value)

        else:
            operand = op_l
            c_val = float(op_r.value)

        _lb, _ub = operand.bounds
        _lb = float(_lb)
        _ub = float(_ub)
        # add Less row (a(value + tolerance - r_LB) + r > value + tolerance) and set mip bounds to
        # -inf;-comparison_val
        # add Less row (a(r_UB - value - tolerance) + r < r_UB) and set mip bounds to lb;inf
        _coefs = [
            {less_op: c_val + ModelConstants.TOLERANCE - _lb, operand.key(): 1.0},
            {less_op: _ub - c_val - ModelConstants.TOLERANCE, operand.key(): 1.0}
        ]
        _lbs = [c_val + ModelConstants.TOLERANCE, self.model_default_lb]
        _ubs = [self.model_default_ub, _ub]

        return ConstraintContainer(name=None, coefs=_coefs, lbs=_lbs, ubs=_ubs)

    @staticmethod
    def equal_constraint(symbolic):
        """
        Following Propositional logic, a predicate (a => r = value) can be translated as: r - value = 0
        :param symbolic: the symbolic expression
        :return: a constraint
        """
        less_op = symbolic.key()
        op_l = symbolic.variables[0]
        op_r = symbolic.variables[1]

        if op_l.is_numeric:
            operand = op_r
            c_val = float(op_l.value)

        else:
            operand = op_l
            c_val = float(op_r.value)

        _coefs = [{less_op: - c_val, operand.key(): 1.0}]
        _lbs = [0]
        _ubs = [0]

        return ConstraintContainer(name=None, coefs=_coefs, lbs=_lbs, ubs=_ubs)

    @staticmethod
    def none_constraint(_):
        """
        The target or reaction can take any boolean value (0;1),
        so a constraint with bounds to 0;1 is added to the problem
        :param _:
        :return:
        """
        return ConstraintContainer(name=None, coefs=[{}], lbs=[0.0], ubs=[1.0])

    @staticmethod
    def false_constraint(_):
        """
        Constraint with 0;0 bounds
        :param _:
        :return:
        """
        return ConstraintContainer(name=None, coefs=[{}], lbs=[0.0], ubs=[0.0])

    @staticmethod
    def true_constraint(_):
        """
        Constraint with 1;1 bounds
        :param _:
        :return:
        """
        return ConstraintContainer(name=None, coefs=[{}], lbs=[1.0], ubs=[1.0])

    @staticmethod
    def number_constraint(symbolic):
        """
        Constraint with number;number bounds
        :param symbolic:
        :return:
        """
        return ConstraintContainer(name=None, coefs=[{}], lbs=[float(symbolic.value)], ubs=[float(symbolic.value)])

    @staticmethod
    def symbol_constraint(symbolic):
        """
        Constraint with -1*symbol and 0;0 bounds
        :param symbolic:
        :return:
        """
        return ConstraintContainer(name=None, coefs=[{symbolic.key(): -1.0}], lbs=[0.0], ubs=[0.0])

    def get_lp_constraint(self,
                          symbolic,
                          operators=True,
                          bool_atoms=True,
                          numeric_atoms=True,
                          symbolic_atoms=True,
                          empty_symbolic=True,
                          ):
        """
        Get the constraint corresponding to the symbolic expression
        :param symbolic: the symbolic expression
        :param operators: if True, the operators constraints are returned
        :param bool_atoms: if True, the boolean atoms constraints are returned
        :param numeric_atoms: if True, the numeric atoms constraints are returned
        :param symbolic_atoms: if True, the symbolic atoms constraints are returned
        :param empty_symbolic: if True, the empty symbolic constraints are returned
        :return: a constraint or None if there is no constraint for the given symbolic expression
        """
        if operators:

            if symbolic.is_and:
                return partial(self.and_constraint, symbolic)

            elif symbolic.is_or:
                return partial(self.or_constraint, symbolic)

            elif symbolic.is_not:
                return partial(self.not_constraint, symbolic)

            elif symbolic.is_greater or symbolic.is_greater_equal:
                return partial(self.greater_constraint, symbolic)

            elif symbolic.is_less or symbolic.is_less_equal:
                return partial(self.less_constraint, symbolic)

            elif symbolic.is_equal:
                return partial(self.equal_constraint, symbolic)

        if bool_atoms:
            if symbolic.is_true:

                return partial(self.true_constraint, symbolic)

            elif symbolic.is_false:

                return partial(self.false_constraint, symbolic)

        if numeric_atoms:
            if symbolic.is_numeric:
                return partial(self.number_constraint, symbolic)

        if symbolic_atoms:
            if symbolic.is_symbol:
                return partial(self.symbol_constraint, symbolic)

        if empty_symbolic:
            if symbolic.is_none:
                return partial(self.none_constraint, symbolic)

        return

    @staticmethod
    def _variable_operator(symbolic):
        name = symbolic.key()
        names = []
        lbs = []
        ubs = []
        var_types = []

        for i, _ in enumerate(symbolic.variables[:-1]):
            names.append(f'{name}_{i}')
            lbs.append(0.0)
            ubs.append(1.0)
            var_types.append(VarType.INTEGER)

        return VariableContainer(name=name, sub_variables=names, lbs=lbs, ubs=ubs, variables_type=var_types)

    def and_variable(self, symbolic):
        """
        The AND operator is translated as a variable with 0;1 bounds
        :param symbolic: the symbolic expression
        :return: a variable
        """
        return self._variable_operator(symbolic=symbolic)

    def or_variable(self, symbolic):
        """
        The OR operator is translated as a variable with 0;1 bounds
        :param symbolic: the symbolic expression
        :return: a variable
        """
        return self._variable_operator(symbolic=symbolic)

    @staticmethod
    def not_variable(symbolic):
        """
        The NOT operator is translated as a variable with 0;1 bounds
        :param symbolic: the symbolic expression
        :return: a variable
        """
        name = symbolic.key()
        sub_variable_name = f'{name}_0'

        return VariableContainer(name=symbolic.key(),
                                 sub_variables=[sub_variable_name],
                                 lbs=[0.0],
                                 ubs=[1.0],
                                 variables_type=[VarType.INTEGER])

    def greater_variable(self, symbolic):
        """
        The greater operator is translated as a variable with 0;1 bounds
        :param symbolic: the symbolic expression
        :return: a variable
        """
        return self._variable_operator(symbolic=symbolic)

    def less_variable(self, symbolic):
        """
        The less operator is translated as a variable with 0;1 bounds
        :param symbolic: the symbolic expression
        :return: a variable
        """
        return self._variable_operator(symbolic=symbolic)

    def equal_variable(self, symbolic):
        """
        The equal operator is translated as a variable with 0;1 bounds
        :param symbolic: the symbolic expression
        :return: a variable
        """
        return self._variable_operator(symbolic=symbolic)

    @staticmethod
    def symbol_variable(symbolic):
        """
        The symbol is translated as a variable with alpha < symbol < beta bounds
        :param symbolic: the symbolic expression
        :return: a variable
        """
        name = symbolic.key()
        lb, ub = symbolic.bounds
        lb = float(lb)
        ub = float(ub)

        if (lb, ub) in integer_coefficients:
            v_type = VarType.INTEGER
        else:
            v_type = VarType.CONTINUOUS

        return VariableContainer(name=name, sub_variables=[name], lbs=[lb], ubs=[ub], variables_type=[v_type])

    def get_lp_variable(self, symbolic):
        """
        Get the variable corresponding to the symbolic expression
        :param symbolic: the symbolic expression
        :return: a variable or None if there is no variable for the given symbolic expression
        """

        if symbolic.is_and:
            return partial(self.and_variable, symbolic)

        elif symbolic.is_or:
            return partial(self.or_variable, symbolic)

        elif symbolic.is_not:
            return partial(self.not_variable, symbolic)

        elif symbolic.is_greater or symbolic.is_greater_equal:
            return partial(self.greater_variable, symbolic)

        elif symbolic.is_less or symbolic.is_less_equal:
            return partial(self.less_variable, symbolic)

        elif symbolic.is_equal:
            return partial(self.equal_variable, symbolic)

        elif symbolic.is_symbol:
            return partial(self.symbol_variable, symbolic)

        return

    def linearize_atomic_expression(self, boolean_variable, symbolic):
        """
        It builds the variables and constraints corresponding to the linearization of the atomic expression
        :param boolean_variable: the boolean variable corresponding to the atomic expression
        :param symbolic: the symbolic expression
        :return: a list of variables and a list of constraints
        """
        variables = []
        constraints = []

        if symbolic.is_symbol:
            var = self.symbol_variable(symbolic=symbolic)
            variables.append(var)

        linearizer = self.get_lp_constraint(symbolic, operators=False)

        expression_cnt = linearizer()

        expression_cnt.coefs[0][boolean_variable] = 1.0

        constraints.append(expression_cnt)

        return variables, constraints

    def linearize_complex_expression(self, boolean_variable, symbolic):
        """
        It builds the variables and constraints corresponding to the linearization of the complex expression

        It iterates an expression in the reverse order
        For example, expr = A and (B or C) yields the following elements:
            - C
            - B
            - A
            - (B or C)
            - A and (B or C)
        For each element, the linearization is performed and the resulting variables and constraints
        are added to the list. To see which variables and constraints are added for each element,
        see the constraint methods:
            - `and_constraint`
            - `or_constraint`
            - `not_constraint`
            - `greater_constraint`
            - `less_constraint`
            - `equal_constraint`
            - `symbol_constraint`
            - `none_constraint`

        and the variable methods:
            - `and_variable`
            - `or_variable`
            - `not_variable`
            - `greater_variable`
            - `less_variable`
            - `equal_variable`
            - `symbol_variable`

        :param boolean_variable: the boolean variable corresponding to the complex expression
        :param symbolic: the symbolic expression
        :return: a list of variables and a list of constraints
        """
        variables = []
        constraints = []
        last_variable = None
        for atom in symbolic:

            # An operator expression will be decomposed into multiple operator expressions, namely into a nested
            # expression. For instance, an A & B & C & D will become And(D, And(C, And(A, B))). Each nested operator
            # expression will be a column in the matrix and then linked in the columns. However, this operator
            # expression will be the same for all maters and have the same column identifier regardless of the length.
            # Thus, all operator columns (variables) will be under the columns linked list engine. When retrieving
            # the indexes of the operator expression, the last index
            # of the slice should be used to get the last real column, as the result of this column is the one that
            # really matters. The hashes of the columns for the simulation engine should be the row name plus
            # str of the operator
            last_variable = atom

            lp_variable = self.get_lp_variable(symbolic=atom)

            if lp_variable is not None:
                var = lp_variable()
                variables.append(var)

            lp_constraint = self.get_lp_constraint(atom,
                                                   bool_atoms=False,
                                                   numeric_atoms=False,
                                                   symbolic_atoms=False,
                                                   empty_symbolic=False)

            if lp_constraint is not None:
                constraint = lp_constraint()
                constraints.append(constraint)

        # identifying the last index to link the outcome of this variable to the boolean variable associated to the
        # expression
        last_variable_name = last_variable.key()
        names = [f'{last_variable_name}_{i}' for i, _ in enumerate(last_variable.variables[:-1])]

        if names:
            last_variable_name = names[-1]
        else:
            last_variable_name = last_variable_name

        # add gene row which means that the gene variable in the mip matrix is associated with the last midterm
        # expression, namely the whole expression
        # set mip bounds to 0;0
        expression_cnt = ConstraintContainer(name=None,
                                             coefs=[{boolean_variable: 1.0, last_variable_name: -1.0}],
                                             lbs=[0.0],
                                             ubs=[0.0])
        constraints.append(expression_cnt)

        return variables, constraints

    def linearize_expression(self, boolean_variable, symbolic):
        """
        It builds the variables and constraints corresponding to the linearization of the expression.

        It iterates an expression in the reverse order
        For example, expr = A and (B or C) yields the following elements:
            - C
            - B
            - A
            - (B or C)
            - A and (B or C)
        For each element, the linearization is performed and the resulting variables and constraints
        are added to the list. To see which variables and constraints are added for each element,
        see the constraint methods:
            - `and_constraint`
            - `or_constraint`
            - `not_constraint`
            - `greater_constraint`
            - `less_constraint`
            - `equal_constraint`
            - `symbol_constraint`
            - `none_constraint`

        and the variable methods:
            - `and_variable`
            - `or_variable`
            - `not_variable`
            - `greater_variable`
            - `less_variable`
            - `equal_variable`
            - `symbol_variable`
        :param boolean_variable: the boolean variable corresponding to the expression
        :param symbolic: the symbolic expression
        :return: a list of variables and a list of constraints
        """
        # if expression is atom and defines a variable always On or Off, add an On/Off row
        if symbolic.is_atom:
            return self.linearize_atomic_expression(boolean_variable=boolean_variable, symbolic=symbolic)

        return self.linearize_complex_expression(boolean_variable=boolean_variable, symbolic=symbolic)

    def _build_interactions(self):
        """
        It builds the algebraic constraints for SRFBA, namely the GPR constraints and the interaction constraints.
        It is called automatically upon instantiation if build is True.
        :return:
        """
        variables = []
        constraints = []
        for interaction in self.model.yield_interactions():
            interaction_variables, interaction_constraints = self.interaction_constraint(interaction)
            variables.extend(interaction_variables)
            constraints.extend(interaction_constraints)

        self.add_variables(*variables)
        self.add_constraints(*constraints)

    def _build_gprs(self):
        """
        It builds the algebraic constraints for SRFBA, namely the GPR constraints and the interaction constraints.
        It is called automatically upon instantiation if build is True.
        :return:
        """
        variables = []
        constraints = []

        for reaction in self.model.yield_reactions():
            gpr_variables, gpr_constraints = self.gpr_constraint(reaction)
            variables.extend(gpr_variables)
            constraints.extend(gpr_constraints)

        self.add_variables(*variables)
        self.add_constraints(*constraints)

    def _build(self):
        """
        It builds the linear problem for SRFBA. It is called automatically upon instantiation if build is True.
        The SRFBA problem is a mixed-integer linear problem (MILP) with the following structure:
            - metabolic constraints
            - GPR constraints
            - interaction constraints

        :return:
        """
        if self.model.is_metabolic() and self.model.is_regulatory():
            self._build_mass_constraints()
            self._build_gprs()
            self._build_interactions()

            self._linear_objective = {var.id: value for var, value in self.model.objective.items()}
            self._minimize = False

    def _optimize(self,
                  to_solver: bool = False,
                  solver_kwargs: Dict = None,
                  initial_state: Dict[str, float] = None,
                  **kwargs) -> Union[ModelSolution, Solution]:
        """
        It solves the linear problem. The linear problem is solved using the solver interface.

        The optimize method allows setting temporary changes to the linear problem. The changes are
        applied to the linear problem reverted to the original state afterward.
        Objective, constraints and solver parameters can be set temporarily.

        The solution is returned as a ModelSolution instance, unless to_solver is True. In this case,
        the solution is returned as a SolverSolution instance.

        :param to_solver: Whether to return the solution as a SolverSolution instance. Default: False
        :param solver_kwargs: A dictionary of solver parameters to be set temporarily. Default: None
        :param initial_state: a dictionary of variable ids and their values to set as initial state
        :return: A ModelSolution instance or a SolverSolution instance if to_solver is True.
        """
        if not initial_state:
            initial_state = {}

        if not solver_kwargs:
            solver_kwargs = {}

        if 'constraints' in solver_kwargs:
            constraints = solver_kwargs['constraints'].copy()

        else:
            constraints = {}

        constraints = {**constraints, **initial_state}
        solver_kwargs['constraints'] = constraints

        solution = self.solver.solve(**solver_kwargs)
        return solution



import os
from functools import partial
from typing import Union, TYPE_CHECKING

from mewpy.io.dto import DataTransferObject, VariableRecord, History, FunctionTerm, CompartmentRecord
from mewpy.germ.algebra import Expression, Symbol, NoneAtom, Float
from mewpy.germ.models import RegulatoryModel, MetabolicModel
from mewpy.util.constants import ModelConstants
from .engine import Engine
from .engines_utils import (ASTNODE_BOOLEAN_VALUES, ASTNODE_RELATIONAL_OPERATORS,
                            ASTNODE_NAME, ASTNODE_BOOLEAN_OPERATORS, ASTNODE_VALUES,
                            f_id, fs_id,
                            F_GENE, F_SPECIE, F_REACTION, F_SPECIE_REV, F_GENE_REV, F_REACTION_REV, F_TRANSITION,
                            F_TRANSITION_REV,
                            get_sbml_doc_to_write, get_sbml_doc_to_read,
                            write_sbml_doc,
                            set_math, expression_warning, sbml_warning)

if TYPE_CHECKING:
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


class RegulatorySBML(Engine):
    def __init__(self, io, config, model=None):

        super().__init__(io, config, model)

    @property
    def model_type(self):
        return 'regulatory'

    @property
    def model(self):

        if self._model is None:

            identifier = self.dto.id

            if not identifier:
                identifier = self.get_identifier()

            return RegulatoryModel(identifier=identifier)

        return self._model

    @staticmethod
    def _parse_coefficients_note(notes):

        if not notes:
            return set()

        coefficients = set()

        for row in notes.split('\n'):

            row = row.strip()

            if row.startswith('<p>') and row.endswith('</p>'):
                _, level = row.split(':')

                # st = st.strip('<p>')
                # level = level.strip('</p>')
                #
                # if 'state' in st.lower():
                #     st = int(st.lower().replace('state', '').strip())

                lb = ModelConstants.REACTION_LOWER_BOUND
                ub = ModelConstants.REACTION_UPPER_BOUND

                level = level.replace('+inf', f'{ub}').replace('-inf', f'{lb}')

                min_coef, max_coef = level.split(',')

                min_coef = float(min_coef[1:])
                coefficients.add(min_coef)

                max_coef = float(max_coef[0:-1])
                coefficients.add(max_coef)

        return coefficients

    @staticmethod
    def _parse_initial_level_note(notes):

        if not notes:
            return

        for row in notes.split('\n'):

            row = row.strip()

            if row.startswith('<p>') and row.endswith('</p>'):
                parameter, level = row.split(':')

                parameter = parameter.strip('<p>')
                level = level.strip('</p>')

                if 'initial' in parameter.lower():
                    return float(level.replace(' ', ''))

    def _update_qual_species_coefficients(self, qual_species, value):

        # all species are created with the following coefficients: {0:0, 1: maximum level}.
        # but intermediate states can be encoded into the transitions and thus added to the coefficients dict

        value = float(value)

        if qual_species not in self.dto.variables:
            raise SyntaxError

        coefficients = self.dto.variables[qual_species].coefficients

        if min(coefficients) > value > max(coefficients):
            raise SyntaxError

        coefficients.add(value)

    def _parse_values(self, ast_node, ast_node_qual_species, inputs_thresholds):

        node_type = ast_node.getType()

        if node_type in ASTNODE_BOOLEAN_VALUES:

            return ASTNODE_BOOLEAN_VALUES[node_type]

        elif ast_node.getType() == ASTNODE_NAME:

            # In this case, the right child is encoded in the list of inputs

            input_id = fs_id(ast_node.getName(), (F_SPECIE, F_GENE, F_REACTION, F_TRANSITION))
            value = inputs_thresholds.get(input_id, None)

            if value is None:
                raise SyntaxError

            regulator_id = fs_id(ast_node_qual_species.getName(), (F_SPECIE, F_GENE, F_REACTION))
            self._update_qual_species_coefficients(regulator_id, value)

            return Float(value)

        elif ast_node.getType() in ASTNODE_VALUES:

            # In this case, the right child is encoded in the regulators current coefficients

            value = ast_node.getValue()

            if value is None:
                raise SyntaxError

            regulator_id = fs_id(ast_node_qual_species.getName(), (F_SPECIE, F_GENE, F_REACTION))
            self._update_qual_species_coefficients(regulator_id, value)

            return Float(value)

        else:
            raise SyntaxError

    def _parse_symbols(self, ast_node):

        ast_node_id = fs_id(ast_node.getName(), (F_SPECIE, F_GENE, F_REACTION))

        if ast_node_id not in self.dto.variables:
            raise SyntaxError

        return Symbol(value=ast_node_id)

    def _parse_math_node(self, ast_node, inputs_thresholds):

        node_type = ast_node.getType()

        if node_type in ASTNODE_BOOLEAN_OPERATORS:

            Operator = ASTNODE_BOOLEAN_OPERATORS[node_type]
            variables = []

            for child in range(ast_node.getNumChildren()):
                variable = self._parse_math_node(ast_node.getChild(child), inputs_thresholds)

                variables.append(variable)

            return Operator(variables=variables)

        elif node_type in ASTNODE_RELATIONAL_OPERATORS:

            Operator = ASTNODE_RELATIONAL_OPERATORS[node_type]
            symbol = self._parse_symbols(ast_node.getChild(0))
            value = self._parse_values(ast_node.getChild(1), ast_node.getChild(0), inputs_thresholds)

            return Operator(variables=[symbol, value])

        elif node_type == ASTNODE_NAME:

            return self._parse_symbols(ast_node)

        else:

            raise SyntaxError

    def parse_math_node(self, ast_node, inputs_thresholds):

        """
        Reads and parses a node of type math ASTNode into an algebraic expression.

        :param ast_node: ASTNode, math ASTNode from functionTerm.getMath()
        :param inputs_thresholds:
        :return: Symbolic
        """

        if not inputs_thresholds:
            inputs_thresholds = {}

        try:

            symbolic = self._parse_math_node(ast_node, inputs_thresholds)

        except SyntaxError:

            self.warnings.append(partial(expression_warning,
                                         f'{ast_node} cannot be parsed. Assigning empty expression instead'))

            symbolic = NoneAtom()

        return symbolic

    def get_identifier(self):

        if os.path.exists(self.io):
            _, identifier = os.path.split(self.io)
            return os.path.splitext(identifier)[0]

        return 'model'

    def _open_to_read(self):

        self._dto = DataTransferObject()

        # -----------------------------------------------------------------------------
        # Doc, Model and Qual Plugin
        # -----------------------------------------------------------------------------

        self.dto.doc = get_sbml_doc_to_read(self.io)

        self.dto.model = self.dto.doc.getModel()

        if self.dto.model is None:
            raise OSError(f'{self.io} is not a valid input. Model SBML section is missing. '
                          f'Provide a correct path or file handler')

        self.dto.qual_plugin = self.dto.model.getPlugin('qual')

        if self.dto.qual_plugin is None:
            raise OSError(f'Although {self.io} is a valid SBML file, the qual plugin was not detected. Thus, '
                          f'regulatory interactions cannot be determined')

        identifier = self.dto.model.getIdAttribute()

        if not identifier:
            self.warnings.append(partial(sbml_warning, 'Model identifier is not encoded in the SBML file'))

            identifier = self.get_identifier()

        self.dto.id = identifier
        self.dto.name = self.dto.model.getName()

    def _open_to_write(self):

        self._dto = DataTransferObject()

        # -----------------------------------------------------------------------------
        # Doc, Model and FBC Plugin
        # -----------------------------------------------------------------------------
        self.dto.doc = get_sbml_doc_to_write(self.io,
                                             level=3,
                                             version=1,
                                             packages=('qual',),
                                             packages_version=(1,),
                                             packages_required=(True,),
                                             sbo_term=False)

        if self.dto.model is None:

            self.dto.model = self.dto.doc.createModel()

        else:

            self.dto.model = self.dto.doc.getModel()

        self.dto.qual_plugin = self.dto.model.getPlugin('qual')

        if self.model.id is not None:
            self.dto.model.setId(self.model.id)
            self.dto.model.setMetaId('meta_' + self.model.id)

        else:
            self.dto.model.setMetaId('meta_model')

        if self.model.name is not None:
            self.dto.model.setName(self.model.name)

    def open(self, mode='r'):

        if mode == 'r':

            return self._open_to_read()

        elif mode == 'w':

            return self._open_to_write()

        else:
            raise ValueError(f'{mode} mode is not recognized. Try one of the following: r, w')

    def parse(self):

        if self.dto is None:
            raise OSError('SBML file is not open')

        if self.dto.id is None:
            raise OSError('SBML file is not open')

        if self.dto.doc is None:
            raise OSError('SBML file is not open')

        if self.dto.model is None:
            raise OSError(f'SBML file is not open')

        if self.dto.qual_plugin is None:
            raise OSError(f'SBML file is not open')

        self.dto.level = self.dto.model.getLevel()
        self.dto.version = self.dto.model.getVersion()

        has_history = self.dto.model.isSetModelHistory()

        if has_history:

            history = self.dto.model.getModelHistory()

            created = None
            if history.isSetCreatedDate():
                created = history.getCreatedDate()

            creators = history.getListCreators()

            self.dto.history = History(data=created, creators=creators)

        # -----------------------------------------------------------------------------
        # Compartments
        # -----------------------------------------------------------------------------
        for compartment in self.dto.model.getListOfCompartments():
            self.dto.compartments[compartment.getIdAttribute()] = CompartmentRecord(id=compartment.getIdAttribute(),
                                                                                    name=compartment.getName())

        # -----------------------------------------------------------------------------
        # Qualitative Species
        # -----------------------------------------------------------------------------
        for qual_species in self.dto.qual_plugin.getListOfQualitativeSpecies():

            identifier = fs_id(qual_species.getIdAttribute(), (F_SPECIE, F_GENE, F_REACTION))
            name = qual_species.getName()

            if not name:
                name = identifier

            aliases = {identifier, name}
            compartment = qual_species.getCompartment()

            constant = qual_species.getConstant()

            # coefficients can be encoded into the notes section. It is not advised though
            notes = qual_species.getNotesString()

            coefficients = self._parse_coefficients_note(notes)

            if coefficients:

                self.warnings.append(partial(sbml_warning, f'Are the {identifier} coefficients encoded in the notes '
                                                           f'section? Coefficients must be hard coded during the '
                                                           f'math nodes of each transition.'))

                maximum_level = max(coefficients)
                minimum_level = min(coefficients)

            else:

                # coefficients might be updated later during transition parsing
                maximum_level = float(qual_species.getMaxLevel())
                minimum_level = 0.0

                coefficients = {minimum_level, maximum_level}

            # finding the default coefficient
            if qual_species.isSetInitialLevel():
                active_coefficient = float(qual_species.getInitialLevel())

            else:

                active_coefficient = self._parse_initial_level_note(notes)

                if active_coefficient is None:
                    self.warnings.append(partial(sbml_warning, f'{identifier} initial level was not found. '
                                                               f'Setting default/initial coefficient to the '
                                                               f'minimum value'))

                    active_coefficient = 0.0

            if minimum_level > active_coefficient > maximum_level:

                raise ValueError(f'Initial level is higher/lower than the minimum/maximum level for the {identifier} '
                                 f'qual species')

            else:

                coefficients.add(active_coefficient)

            variable = VariableRecord(id=identifier,
                                      name=name,
                                      aliases=aliases,
                                      compartment=compartment,
                                      constant=constant,
                                      notes=notes,
                                      coefficients=coefficients,
                                      active_coefficient=active_coefficient)

            self.dto.variables[identifier] = variable

        # -----------------------------------------------------------------------------
        # Interactions/List of Transitions
        # -----------------------------------------------------------------------------

        # iterate over all transitions/interactions
        for transition in self.dto.qual_plugin.getListOfTransitions():

            identifier = f_id(transition.getIdAttribute(), F_TRANSITION)
            name = transition.getName()

            if not name:
                name = identifier

            # -----------------------------------------------------------------------------
            # Regulators/List of Inputs
            # -----------------------------------------------------------------------------

            # regulators can be encoded into the list of inputs by the qualitative species or by an input identifier.
            # A regulator associated with an input identifier must also be associated with the qualitative species
            # and sometimes with the threshold level.
            # The threshold level stands for the coefficient of the regulator in a given function term.
            # In this case, the ASTNode math parsing will have to replace the input id the is enclosed by the tag <ci>
            # by the threshold level.

            # A regulator encoded into the list of inputs using only the qualitative species is a simpler approach.
            # In this case, the threshold level is regularly encoded in a <cn> tag of type integer or rational

            # Nevertheless, the second option is more powerful, as we can set multiple non-integer values

            # to be used during functional term parsing
            regulators = {}
            # to be used during functional term parsing
            inputs_thresholds = {}

            for regulator_input in transition.getListOfInputs():

                if not regulator_input.isSetQualitativeSpecies():
                    raise ValueError(f"Qualitative species for {identifier}'s transition input is not set")

                regulator_id = fs_id(regulator_input.getQualitativeSpecies(), (F_SPECIE, F_GENE, F_REACTION))

                regulator_record = self.dto.variables[regulator_id]

                regulators[regulator_id] = regulator_record

                self.dto.regulators[regulator_id] = regulator_record

                self.variables[regulator_id].add('regulator')

                # input identifier
                if regulator_input.isSetId():

                    input_id = fs_id(regulator_input.getIdAttribute(), (F_SPECIE, F_GENE, F_REACTION, F_TRANSITION))

                else:
                    input_id = regulator_id

                # input threshold level
                if regulator_input.isSetThresholdLevel():
                    input_threshold = regulator_input.getThresholdLevel()

                    inputs_thresholds[input_id] = input_threshold

            if inputs_thresholds:
                self.warnings.append(partial(sbml_warning, f'{identifier} threshold levels detected. It is '
                                                           f'recommended to encode input levels directly in the math '
                                                           f'node'))

            # -----------------------------------------------------------------------------
            # Targets/List of Outputs
            # -----------------------------------------------------------------------------

            # In the case of outputs/targets list, only the qualitative species really matters
            targets = {}
            for target_output in transition.getListOfOutputs():

                if not target_output.isSetQualitativeSpecies():
                    raise ValueError(f"Qualitative species for {identifier}'s transition output is not set")

                target_id = fs_id(target_output.getQualitativeSpecies(), (F_SPECIE, F_GENE, F_REACTION))

                target_record = self.dto.variables[target_id]

                targets[target_id] = target_record

                self.dto.targets[target_id] = target_record

                self.variables[target_id].add('target')

            # -----------------------------------------------------------------------------
            # Function Terms
            # -----------------------------------------------------------------------------

            # A function term specify how regulators determine the state of the targets
            function_terms = {}

            # -----------------------------------------------------------------------------
            # Default term
            # -----------------------------------------------------------------------------

            # The default term does not have an expression, as it just refers to the base coefficient value
            # for the target.
            default_term = transition.getDefaultTerm()

            if default_term is None:

                self.warnings.append(partial(sbml_warning, f'Default function term '
                                                           f'not set for transition {identifier}.'
                                                           f'Setting default function term of zero.'))

                result_level = 0.0

            else:

                result_level = float(default_term.getResultLevel())

                if result_level is None:
                    result_level = 0.0

                    self.warnings.append(partial(sbml_warning, 'Default function term result level not found. '
                                                               'Setting to zero'))

            function_terms[result_level] = FunctionTerm(id=f'{identifier}_{result_level}',
                                                        symbolic=NoneAtom(),
                                                        coefficient=result_level)

            # -----------------------------------------------------------------------------
            # Math based function terms
            # -----------------------------------------------------------------------------

            # The other function terms define how the outputs/targets can take multiple states.
            # These terms encode the conditions and logic for a given target being active or inactive
            for function_term in transition.getListOfFunctionTerms():

                # result level of the function term
                result_level = function_term.getResultLevel()

                if result_level is None:
                    result_level = 0

                    self.warnings.append(partial(sbml_warning, f'Function term result level not found '
                                                               f'for term in transition {identifier}. '
                                                               f'Setting to zero'))

                # a transition can only have one function term per result level
                if result_level in function_terms:
                    raise ValueError(f'Function term {identifier} with {result_level} result level has already '
                                     f'been used by other function terms')

                # a transition can only have function terms to which the result levels are less or equal than the
                # maximum levels of all its outputs. _update_qual_species_coefficients will check these issues
                for target in targets:
                    self._update_qual_species_coefficients(target, result_level)

                # math node of the function term
                if function_term.isSetMath():

                    math_node = function_term.getMath()

                    # It will parse the symbolic algebraic expression in the ASTMath node. The numeric atoms will be
                    # replaced by the state coefficient according to the inputs' threshold levels or qualitative
                    # species' coefficients
                    symbolic = self.parse_math_node(math_node, inputs_thresholds)

                else:

                    self.warnings.append(partial(sbml_warning, f'Function term math node was'
                                                               f'not set for transition {identifier} with '
                                                               f'{result_level}.'
                                                               f'Setting function term equal to the result level.'))

                    symbolic = NoneAtom()

                function_terms[result_level] = FunctionTerm(id=f'{identifier}_{result_level}',
                                                            symbolic=symbolic,
                                                            coefficient=result_level)

            # -----------------------------------------------------------------------------
            # Interaction record
            # -----------------------------------------------------------------------------

            # Setting an interaction per target/output

            for target in targets.values():
                interaction_id = f'{target.id}_interaction'

                interaction_record = VariableRecord(id=interaction_id,
                                                    name=identifier,
                                                    aliases={interaction_id, identifier, name, target.id},
                                                    target=target,
                                                    function_terms=function_terms,
                                                    regulators=regulators)

                self.dto.interactions[interaction_id] = interaction_record

                self.variables[interaction_id].add('interaction')

    def read(self,
             model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
             variables=None):

        if not model:
            model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = self.model

        if not variables:
            variables = self.variables

        if self.dto.id:
            model._id = self.dto.id

        if self.dto.name:
            model.name = self.dto.name

        model.compartments = {compartment.id: compartment.name
                              for compartment in self.dto.compartments.values()}

        processed_vars = set()

        for interaction_id, interaction_record in self.dto.interactions.items():

            target_record = interaction_record.target

            target, warning = target_record.to_variable(model=model,
                                                        types=variables.get(target_record.id, {'target'}),
                                                        name=target_record.name,
                                                        aliases=target_record.aliases,
                                                        coefficients=target_record.coefficients)

            if warning:
                self.warnings.append(partial(sbml_warning, warning))

            processed_vars.add(target_record.id)

            regulators_records = interaction_record.regulators

            regulators = {}

            for regulator_id, regulator_record in regulators_records.items():

                regulator, warn = regulator_record.to_variable(model=model,
                                                               types=variables.get(regulator_id, {'regulator'}),
                                                               name=regulator_record.name,
                                                               aliases=regulator_record.aliases,
                                                               coefficients=regulator_record.coefficients)

                if warn:
                    self.warnings.append(partial(sbml_warning, warn))

                regulators[regulator_id] = regulator

                processed_vars.add(regulator_id)

            regulatory_events = {}

            for func_term in interaction_record.function_terms.values():
                expression_regulators = {symbol.name: regulators[symbol.name]
                                         for symbol in func_term.symbolic.atoms(symbols_only=True)}

                regulatory_events[func_term.coefficient] = Expression(symbolic=func_term.symbolic,
                                                                      variables=expression_regulators)

            interaction, warn = interaction_record.to_variable(model=model,
                                                               types=variables.get(interaction_id, {'interaction'}),
                                                               name=interaction_record.name,
                                                               aliases=interaction_record.aliases,
                                                               target=target,
                                                               regulatory_events=regulatory_events)

            if warn:
                self.warnings.append(partial(sbml_warning, warn))

            model.add(interaction, comprehensive=True)

        if len(processed_vars) != len(self.dto.variables):

            for variable_id, variable_record in self.dto.variables.items():

                if variable_id not in processed_vars:

                    variable, warn = variable_record.to_variable(model=model,
                                                                 types=variables.get(variable_id, {'regulator'}),
                                                                 name=variable_record.name,
                                                                 aliases=variable_record.aliases,
                                                                 coefficients=variable_record.coefficients)

                    if warn:
                        self.warnings.append(partial(sbml_warning, warn))

                    model.add(variable)

        return model

    def _reverse_f_id(self, variable):

        if variable.is_reaction():

            return f_id(variable.id, F_REACTION_REV)

        elif variable.is_metabolite():

            return f_id(variable.id, F_SPECIE_REV)

        elif variable.is_gene():

            return f_id(variable.id, F_GENE_REV)

        elif variable.is_interaction():

            if variable.target:

                target_id = self._reverse_f_id(variable.target)

                return f_id(target_id, F_TRANSITION_REV)

            else:

                return f_id(variable.id, F_TRANSITION_REV)

        else:

            return variable.id

    @staticmethod
    def _expression_replace(expression_string, replacements):

        for key, value in replacements.items():
            expression_string = expression_string.replace(key, value)

        return expression_string

    def write(self):

        if self.dto is None:
            raise OSError('SBML file is not open')

        if self.dto.doc is None:
            raise OSError('SBML file is not open')

        if self.dto.model is None:
            raise OSError(f'SBML file is not open')

        # -----------------------------------------------------------------------------
        # Compartments
        # -----------------------------------------------------------------------------
        default_compartment = None

        for compartment_id, compartment_name in self.model.compartments.items():
            sbml_compartment = self.dto.model.createCompartment()
            sbml_compartment.setId(compartment_id)
            sbml_compartment.setName(compartment_name)
            sbml_compartment.setConstant(True)

            if default_compartment is None:
                default_compartment = compartment_id

        if default_compartment is None:
            default_compartment = 'e'
            sbml_compartment = self.dto.model.createCompartment()
            sbml_compartment.setId('e')
            sbml_compartment.setName('extracellular')
            sbml_compartment.setConstant(True)

        # -----------------------------------------------------------------------------
        # Variables/Qualitative Species - Targets
        # -----------------------------------------------------------------------------
        processed_regulators = []

        for target in self.model.yield_targets():

            qual_species = self.dto.qual_plugin.createQualitativeSpecies()

            target_id = self._reverse_f_id(target)
            qual_species.setId(target_id)

            if hasattr(target, 'compartment'):

                if target.compartment is None:
                    compartment = default_compartment

                else:
                    compartment = target.compartment

            else:
                compartment = default_compartment

            qual_species.setCompartment(compartment)

            qual_species.setConstant(False)
            qual_species.setName(target.name)

            qual_species.setInitialLevel(int(round(min(target.coefficients))))
            qual_species.setMaxLevel(int(round(max(target.coefficients))))

            processed_regulators.append(target.id)

        # -----------------------------------------------------------------------------
        # Variables/Qualitative Species - Regulators
        # -----------------------------------------------------------------------------

        for regulator in self.model.yield_regulators():

            if regulator.id not in processed_regulators:

                qual_species = self.dto.qual_plugin.createQualitativeSpecies()

                regulator_id = self._reverse_f_id(regulator)
                qual_species.setId(regulator_id)

                if hasattr(regulator, 'compartment'):

                    if regulator.compartment is None:
                        compartment = default_compartment

                    else:
                        compartment = regulator.compartment

                else:
                    compartment = default_compartment

                qual_species.setCompartment(compartment)

                qual_species.setConstant(False)
                qual_species.setName(regulator.name)

                qual_species.setInitialLevel(int(round(min(regulator.coefficients))))
                qual_species.setMaxLevel(int(round(max(regulator.coefficients))))

        # -----------------------------------------------------------------------------
        # Transitions - Interactions
        # -----------------------------------------------------------------------------
        for interaction in self.model.yield_interactions():

            transition = self.dto.qual_plugin.createTransition()

            transition_id = self._reverse_f_id(interaction)

            transition.setId(transition_id)
            transition.setName(interaction.name)

            # -----------------------------------------------------------------------------
            # Output
            # -----------------------------------------------------------------------------

            if interaction.target:
                output = transition.createOutput()

                target_id = self._reverse_f_id(interaction.target)

                output.setId(f'{target_id}_out')
                output.setQualitativeSpecies(target_id)

            # -----------------------------------------------------------------------------
            # Inputs
            # -----------------------------------------------------------------------------

            for regulator in interaction.yield_regulators():
                regulator_id = self._reverse_f_id(regulator)

                input_id = f'{regulator_id}_input'
                reg_input = transition.createInput()
                reg_input.setId(input_id)
                reg_input.setQualitativeSpecies(regulator_id)

            # -----------------------------------------------------------------------------
            # Default Term
            # -----------------------------------------------------------------------------
            default_term = transition.createDefaultTerm()
            default_assigned = False

            for coefficient, expression in interaction.regulatory_events.items():

                # default term assignment
                if expression.is_none and not default_assigned:
                    default_term.setResultLevel(int(coefficient))

                    default_assigned = True

                    continue

                # -----------------------------------------------------------------------------
                # Function Terms
                # -----------------------------------------------------------------------------
                function_term = transition.createFunctionTerm()
                function_term.setResultLevel(int(coefficient))

                # operators to be replaced
                replacements = {'&': '&&',
                                '|': '||',
                                '~': '!',
                                '=': '==',
                                '<==': '<=',
                                '>==': '>='}

                # reverse ids for the variables that must be replaced
                symbols_reverse_ids = {variable.id: self._reverse_f_id(variable)
                                       for variable in expression.variables.values()}

                replacements.update(symbols_reverse_ids)

                expression_string = expression.to_string()

                expression_string = self._expression_replace(expression_string=expression_string,
                                                             replacements=replacements)

                if not expression_string:
                    self.warnings.append(partial(sbml_warning, f'Empty expression to be set as a function term in '
                                                               f'transition {transition_id}'))

                    continue

                warning = set_math(transition_id, expression_string, function_term)

                if warning:
                    self.warnings.append(partial(sbml_warning, warning))

        write_sbml_doc(self.io, self.dto.doc)

    def close(self):

        if hasattr(self.io, 'close'):
            self.io.close()

    def clean(self):
        self._dto = None



from typing import Any, Dict, TYPE_CHECKING, Generator, Tuple, Sequence, Union

from mewpy.util.constants import ModelConstants
from mewpy.util.history import recorder
from mewpy.germ.models.serialization import serialize
from mewpy.util.utilities import generator
from .variable import Variable
from .variables_utils import coefficients_setter

if TYPE_CHECKING:
    from .interaction import Interaction
    from .target import Target


class Regulator(Variable, variable_type='regulator', register=True, constructor=True, checker=True):

    def __init__(self,
                 identifier: Any,
                 coefficients: Sequence[float] = None,
                 interactions: Dict[str, 'Interaction'] = None,
                 **kwargs):
        """
        A regulator is commonly associated with interactions and
        can usually be available as metabolite or reaction or target too.

        Regulators can control the gene expression of several target genes.

        It holds information regarding the coefficients that can take and the interactions to which is associated

        :param identifier: identifier, e.g. b0001
        :param coefficients: the set of coefficients that this regulator can take.
        These coefficients can be expanded later. 0 and 1 are added by default
        :param interactions: the dictionary of interactions to which the regulator is associated with
        """

        # the coefficient initializer sets minimum and maximum coefficients of 0.0 and 1.0
        if not coefficients:
            coefficients = (0.0, 1.0)
        else:
            coefficients = tuple(coefficients)

        if not interactions:
            interactions = {}

        self._coefficients = coefficients
        self._interactions = interactions

        super().__init__(identifier, **kwargs)

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------

    @property
    def types(self):

        # noinspection PyUnresolvedReferences
        _types = {Regulator.variable_type}

        _types.update(super(Regulator, self).types)

        return _types

    def __str__(self):

        if self.is_reaction() or self.is_metabolite():
            return super(Regulator, self).__str__()

        return f'{self.id} || {self.coefficients}'

    def _regulator_to_html(self):
        """
        It returns a html representation.
        """
        html_dict = {'Coefficients': self.coefficients,
                     'Active': self.is_active,
                     'Interactions': ', '.join(self.interactions),
                     'Targets': ', '.join(self.targets),
                     'Environmental stimulus': self.environmental_stimulus}
        return html_dict

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('coefficients', 'coefficients', '_coefficients')
    @property
    def coefficients(self) -> Tuple[float, ...]:
        """
        The coefficient of the regulator
        :return: the coefficient
        """
        if hasattr(self, '_bounds'):

            # if it is a reaction, bounds must be returned
            return self._bounds

        # if it is a metabolite, the bounds coefficient of the exchange reaction must be returned
        elif hasattr(self, 'exchange_reaction'):

            if hasattr(self.exchange_reaction, '_bounds'):
                # noinspection PyProtectedMember
                return self.exchange_reaction._bounds

        return self._coefficients

    @serialize('interactions', 'interactions', '_interactions')
    @property
    def interactions(self) -> Dict[str, 'Interaction']:
        """
        The dictionary of interactions to which the regulator is associated with
        :return: the dictionary of interactions
        """
        return self._interactions.copy()

    @property
    def is_active(self):
        """
        It checks whether the gene is active or not
        :return: True if the gene is active, False otherwise
        """
        return max(self.coefficients) > ModelConstants.TOLERANCE

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------
    @coefficients.setter
    @recorder
    def coefficients(self, value: Union[float, Sequence[float]]):
        """
        The target coefficients setter
        :param value: The target coefficients
        :return:
        """
        coefficients_setter(self, value)

    @interactions.setter
    def interactions(self, value: Dict[str, 'Interaction']):
        """
        The dictionary of interactions to set to the regulator

        It does not perform any check or action on the interactions and targets.
        :param value: the dictionary of interactions
        :return:
        """
        if not value:
            value = {}

        self._interactions = value

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def targets(self) -> Dict[str, 'Target']:
        """
        The dictionary of targets to which the regulator is associated with
        :return: the dictionary of targets
        """
        return {interaction.target.id: interaction.target for interaction in self.yield_interactions()
                if interaction.target is not None}

    @property
    def environmental_stimulus(self) -> bool:
        """
        True if the regulator is an environmental stimulus
        :return: True if the regulator is an environmental stimulus
        """
        if self.types == {'regulator'}:
            return True

        return False

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------

    def yield_interactions(self) -> Generator['Interaction', None, None]:
        """
        Yields the interactions to which the regulator is associated with
        :return: the generator of interactions
        """
        return generator(self._interactions)

    def yield_targets(self) -> Generator['Target', None, None]:
        """
        Yields the targets to which the regulator is associated with
        :return: the generator of targets
        """
        return generator(self.targets)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def ko(self, minimum_coefficient: float = 0.0, history=True):
        """
        Knock out the regulator by setting the minimum coefficient
        :param minimum_coefficient: the minimum coefficient
        :param history: if True, the previous coefficient is stored in the history
        """
        old_coef = tuple(self.coefficients)

        coefficients_setter(self, minimum_coefficient)

        if history:
            self.history.queue_command(undo_func=coefficients_setter,
                                       undo_kwargs={'instance': self,
                                                    'value': old_coef},
                                       func=self.ko,
                                       kwargs={'minimum_coefficient': minimum_coefficient,
                                               'history': False})

    def update(self,
               coefficients: Sequence[float] = None,
               interactions: Dict[str, 'Interaction'] = None,
               **kwargs):
        """
        It updates the regulator
        Note that, some update operations are not registered in history.
        It is strongly advisable to use update outside history context manager
        :param coefficients: the set of coefficients that this regulator can take.
        These coefficients can be expanded later. 0 and 1 are added by default
        :param interactions: the dictionary of interactions to which the regulator is associated with
        :param kwargs: the keyword arguments to pass to the super class
        """
        super().update(**kwargs)

        if coefficients is not None:
            self.coefficients = coefficients

        if interactions is not None:
            self._interactions.update(interactions)



from typing import Any, TYPE_CHECKING, Union, Generator, Dict, List, Tuple, Set

from .model import Model
from mewpy.util.history import recorder
from mewpy.germ.models.serialization import serialize
from mewpy.util.utilities import generator

if TYPE_CHECKING:
    from mewpy.germ.variables import Interaction, Regulator, Target, Metabolite, Reaction


class RegulatoryModel(Model, model_type='regulatory', register=True, constructor=True, checker=True):
    """
    A germ regulatory model can represent a Transcriptional Regulatory Network (TRN), containing
    interactions between regulators and targets.
    However, this regulatory model is not a standard directed TRN graph, but a set of
    interactions, each one containing a single target and a set of regulators.
    Each interaction contains a set of regulatory events that determine the state or coefficient of the target gene.
    A regulatory event consists of a boolean algebra expression and the corresponding coefficient.
    If the boolean expression is evaluated to True, the resulting coefficient is applied to the target gene.
    Otherwise, the coefficient is ignored.

    The model also contains a set of environmental stimuli, which are associated with
    interactions. The environmental stimuli can be used to define the initial state of the model.

    The regulatory model can be loaded with compartments, although these can be inferred from the available
    regulators.

    The regulatory model, as with other models, provides a clean interface for manipulation with the add, remove and
    update methods. One can perform the following operations:
        - Add a new interaction, regulator or target
        - Remove an interaction, regulator or target
        - Update the compartments of the model
    """
    def __init__(self,
                 identifier: Any,
                 compartments: Dict[str, str] = None,
                 interactions: Dict[str, 'Interaction'] = None,
                 regulators: Dict[str, 'Regulator'] = None,
                 targets: Dict[str, 'Target'] = None,
                 **kwargs):

        """
        A germ regulatory model can represent a Transcriptional Regulatory Network (TRN), containing
        interactions between regulators and targets.
        However, this regulatory model is not a standard directed TRN graph, but a set of
        interactions, each one containing a single target and a set of regulators.
        Each interaction contains a set of regulatory events that determine the state or coefficient of the target gene.
        A regulatory event consists of a boolean algebra expression and the corresponding coefficient.
        If the boolean expression is evaluated to True, the resulting coefficient is applied to the target gene.
        Otherwise, the coefficient is ignored.

        The model also contains a set of environmental stimuli, which are associated with
        interactions. The environmental stimuli can be used to define the initial state of the model.

        The regulatory model can be loaded with compartments, although these can be inferred from the available
        regulators.

        The regulatory model, as with other models, provides a clean interface for manipulation with the add, remove and
        update methods. One can perform the following operations:
            - Add a new interaction, regulator or target
            - Remove an interaction, regulator or target
            - Update the compartments of the model

        :param identifier: identifier, e.g. lac_operon_regulation
        :param compartments: a dictionary with additional compartments not encoded in the regulators
        :param interactions: a dictionary with Interaction objects. See variables.Interaction for more info
        :param regulators: a dictionary with Regulator objects. See variables.Regulator for more info
        :param targets: a dictionary with Target objects. See variables.Target for more info
        """
        # compartments attribute can be shared across the children, thus name mangling
        self.__compartments = {}
        self._interactions = {}
        self._regulators = {}
        self._targets = {}

        super().__init__(identifier,
                         **kwargs)

        # the setters will handle adding and removing variables to the correct containers
        self.compartments = compartments
        self.interactions = interactions
        self.regulators = regulators
        self.targets = targets

    # -----------------------------------------------------------------------------
    # Model type manager
    # -----------------------------------------------------------------------------
    @serialize('types', None)
    @property
    def types(self):
        """
        Returns the model types
        :return: a list with the model types
        """
        _types = {RegulatoryModel.model_type}

        _types.update(super(RegulatoryModel, self).types)

        return _types

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('interactions', 'interactions', '_interactions')
    @property
    def interactions(self) -> Dict[str, 'Interaction']:
        """
        It returns a dictionary with the interactions of the model. The keys are the identifiers of the interactions
        and the values are the `Interaction` objects. To retrieve an iterator with the interactions, use the
        `yield_interactions` method.
        Note that the interactions attribute retrieves a copy of the interactions' container.
        To update the interactions container set new `interactions` or use `add` and `remove` methods.
        :return: a dictionary with the interactions of the model
        """
        return self._interactions.copy()

    @serialize('regulators', 'regulators', '_regulators')
    @property
    def regulators(self) -> Dict[str, 'Regulator']:
        """
        It returns a dictionary with the regulators of the model. The keys are the identifiers of the regulators
        and the values are the `Regulator` objects. To retrieve an iterator with the regulators, use the
        `yield_regulators` method.
        Note that the regulators attribute retrieves a copy of the regulators' container.
        To update the regulators container set new `regulators` or use `add` and `remove` methods.
        :return: a dictionary with the regulators of the model
        """
        return self._regulators.copy()

    @serialize('targets', 'targets', '_targets')
    @property
    def targets(self) -> Dict[str, 'Target']:
        """
        It returns a dictionary with the targets of the model. The keys are the identifiers of the targets
        and the values are the `Target` objects. To retrieve an iterator with the targets, use the
        `yield_targets` method.
        Note that the targets attribute retrieves a copy of the targets' container.
        To update the targets container set new `targets` or use `add` and `remove` methods.
        :return: a dictionary with the targets of the model
        """
        return self._targets.copy()

    @property
    def compartments(self) -> Dict[str, str]:
        """
        It returns a dictionary with the compartments of the model. The keys are the identifiers of the compartments
        and the values are the compartment names. To retrieve an iterator with the compartments, use the
        `yield_compartments` method.
        Note that the compartments attribute retrieves a copy of the compartments' container.
        To update the compartments container set new `compartments`.
        :return: a dictionary with the compartments of the model
        """
        compartments = {regulator.compartment: self.__compartments.get(regulator.compartment, '')
                        for regulator in self.yield_regulators()
                        if regulator.is_metabolite() and regulator.compartment is not None}

        compartments.update(self.__compartments)

        compartments.update(super(RegulatoryModel, self).compartments)

        return compartments

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------
    @compartments.setter
    @recorder
    def compartments(self, value: Dict[str, str]):
        """
        It sets the compartments of the model. The keys are the identifiers of the compartments
        and the values are the compartment names.
        :param value: a dictionary with the compartments of the model
        :return:
        """
        if not value:
            value = {}

        self.__compartments.update(value)

    @interactions.setter
    @recorder
    def interactions(self, value: Dict[str, 'Interaction']):
        """
        It sets the interactions of the model. The keys are the identifiers of the interactions
        and the values are the `Interaction` objects.
        :param value: a dictionary with the interactions of the model
        :return:
        """

        if not value:
            value = {}

        self.remove(*self.yield_interactions(), history=False)
        self.add(*value.values(), history=False)

    @regulators.setter
    @recorder
    def regulators(self, value: Dict[str, 'Regulator']):
        """
        It sets the regulators of the model. The keys are the identifiers of the regulators
        and the values are the `Regulator` objects.
        :param value: a dictionary with the regulators of the model
        :return:
        """
        if not value:
            value = {}

        self.remove(*self.yield_regulators(), history=False)
        self.add(*value.values(), history=False)

    @targets.setter
    @recorder
    def targets(self, value: Dict[str, 'Target']):
        """
        It sets the targets of the model. The keys are the identifiers of the targets
        and the values are the `Target` objects.
        :param value: a dictionary with the targets of the model
        :return:
        """
        if not value:
            value = {}

        self.remove(*self.yield_targets(), history=False)
        self.add(*value.values(), history=False)

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------
    @property
    def environmental_stimuli(self) -> Dict[str, 'Regulator']:
        """
        It returns a dictionary with the environmental stimuli of the model. The keys are the identifiers of the
        environmental stimuli and the values are the `Regulator` objects. To retrieve an iterator with the environmental
        stimuli, use the `yield_environmental_stimuli` method.
        :return: a dictionary with the environmental stimuli of the model
        """
        return {reg_id: regulator for reg_id, regulator in self.regulators.items()
                if regulator.environmental_stimulus}

    @property
    def regulatory_reactions(self) -> Dict[str, 'Regulator']:
        """
        It returns a dictionary with the regulatory reactions of the model. The keys are the identifiers of the
        regulatory reactions and the values are the `Regulator` objects. To retrieve an iterator with the regulatory
        reactions, use the `yield_regulatory_reactions` method.
        :return: a dictionary with the regulatory reactions of the model
        """
        return {reg_id: regulator for reg_id, regulator in self.regulators.items()
                if regulator.is_reaction()}

    @property
    def regulatory_metabolites(self) -> Dict[str, 'Regulator']:
        """
        It returns a dictionary with the regulatory metabolites of the model. The keys are the identifiers of the
        regulatory metabolites and the values are the `Regulator` objects. To retrieve an iterator with the regulatory
        metabolites, use the `yield_regulatory_metabolites` method.
        :return: a dictionary with the regulatory metabolites of the model
        """
        return {reg_id: regulator for reg_id, regulator in self.regulators.items()
                if regulator.is_metabolite()}

    # -----------------------------------------------------------------------------
    # Generators
    # -----------------------------------------------------------------------------
    def yield_environmental_stimuli(self) -> Generator['Regulator', None, None]:
        """
        It returns an iterator with the environmental stimuli of the model.
        :return: a generator with the environmental stimuli of the model
        """
        return generator(self.environmental_stimuli)

    def yield_regulatory_reactions(self) -> Generator['Regulator', None, None]:
        """
        It returns an iterator with the regulatory reactions of the model.
        :return: a generator with the regulatory reactions of the model
        """
        return generator(self.regulatory_reactions)

    def yield_regulatory_metabolites(self) -> Generator['Regulator', None, None]:
        """
        It returns an iterator with the regulatory metabolites of the model.
        :return: a generator with the regulatory metabolites of the model
        """
        return generator(self.regulatory_metabolites)

    def yield_interactions(self) -> Generator['Interaction', None, None]:
        """
        It returns an iterator with the interactions of the model.
        :return: a generator with the interactions of the model
        """
        return generator(self._interactions)

    def yield_regulators(self) -> Generator[Union['Regulator', 'Metabolite', 'Reaction'], None, None]:
        """
        It returns an iterator with the regulators of the model.
        :return: a generator with the regulators of the model
        """
        return generator(self._regulators)

    def yield_targets(self) -> Generator['Target', None, None]:
        """
        It returns an iterator with the targets of the model.
        :return: a generator with the targets of the model
        """
        return generator(self._targets)

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------
    def get(self, identifier: Any, default=None) -> Union['Interaction', 'Regulator', 'Target']:
        """
        It returns the object with the given identifier. If the object is not found, it returns the default value.
        For regulatory models, the identifier can be a gene, an interaction or a regulator.
        :param identifier: the identifier of the object
        :param default: the default value to return if the object is not found
        :return: the object associated the given identifier
        """
        if identifier in self._targets:
            return self._targets[identifier]

        elif identifier in self._regulators:
            return self._regulators[identifier]

        elif identifier in self._interactions:
            return self._interactions[identifier]

        else:
            return super(RegulatoryModel, self).get(identifier=identifier, default=default)

    def add(self,
            *variables: Union['Interaction', 'Regulator', 'Target'],
            comprehensive: bool = True,
            history: bool = True):
        """
        It adds the given variables to the model.
        This method accepts a single variable or a list of variables to be added to specific containers in the model.
        The containers to which the variables will be added are specified by the types.

        For instance, if a variable is simultaneously a metabolite and regulator,
        it will be added to the metabolites and regulators containers.

        If comprehensive is True, the variables and their related variables will be added to the model too.
        If history is True, the changes will be recorded in the history.

        This method notifies all simulators with the recent changes.

        :param variables: the variables to be added to the model
        :param comprehensive: if True, the variables and their related variables will be added to the model too
        :param history: if True, the changes will be recorded in the history
        :return:
        """
        if self.is_a('regulatory'):

            for variable in variables:

                if 'target' in variable.types:
                    self._add_variable_to_container(variable, '_targets')

                if 'regulator' in variable.types:
                    self._add_variable_to_container(variable, '_regulators')

                if 'interaction' in variable.types:
                    if comprehensive:

                        if variable.target is not None:
                            self._add_variable_to_container(variable.target, '_targets')

                        for regulator in variable.yield_regulators():
                            self._add_variable_to_container(regulator, '_regulators')

                    self._add_variable_to_container(variable, '_interactions')

        return super(RegulatoryModel, self).add(*variables, comprehensive=comprehensive, history=history)

    def remove(self,
               *variables: Union['Interaction', 'Regulator', 'Target'],
               remove_orphans: bool = False,
               history: bool = True):
        """
        It removes the given variables from the model.
        This method accepts a single variable or a list of variables to be removed from specific containers
        in the model.
        The containers from which the variables will be removed are specified by the types.

        For instance, if a variable is simultaneously a metabolite and regulator,
        it will be removed from the metabolites and regulators containers.

        If remove_orphans is True, the variables and their related variables will be removed from the model too.
        If history is True, the changes will be recorded in the history.

        This method notifies all simulators with the recent changes.

        :param variables: the variables to be removed from the model
        :param remove_orphans: if True, the variables and their related variables will be removed from the model too
        :param history: if True, the changes will be recorded in the history
        :return:
        """
        if self.is_a('regulatory'):

            interactions = set()

            for variable in variables:

                if 'target' in variable.types:
                    self._remove_variable_from_container(variable, '_targets')

                if 'regulator' in variable.types:
                    self._remove_variable_from_container(variable, '_regulators')

                if 'interaction' in variable.types:
                    self._remove_variable_from_container(variable, '_interactions')
                    interactions.add(variable)

            if remove_orphans:
                for interaction in interactions:
                    if interaction.target:
                        self._remove_variable_from_container(interaction.target, '_targets')

                orphan_regulators = self._get_orphans(to_remove=interactions,
                                                      first_container='regulators',
                                                      second_container='interactions')

                for regulator in orphan_regulators:
                    self._remove_variable_from_container(regulator, '_regulators')

        return super(RegulatoryModel, self).remove(*variables, remove_orphans=remove_orphans, history=history)

    def update(self,
               compartments: Dict[str, str] = None,
               variables: Union[List[Union['Interaction', 'Regulator', 'Target']],
                                Tuple[Union['Interaction', 'Regulator', 'Target']],
                                Set[Union['Interaction', 'Regulator', 'Target']]] = None,
               **kwargs):
        """
        It updates the model with relevant information, namely the compartments and the variables.

        :param compartments: the compartments to be added to the model
        :param variables: the variables to be added to the model
        :param kwargs: additional arguments
        :return:
        """
        if compartments is not None:
            self.compartments = compartments

        if variables is not None:
            self.add(*variables)

        super(RegulatoryModel, self).update(**kwargs)



from ..util.constants import EAConstants
from .evaluation import *

engines = dict()


def check_engines():
    global engines
    try:
        from .inspyred.ea import EA as InspyredEA
        engines['inspyred'] = InspyredEA
    except ImportError as e:
        print("inspyred not available")
        print(e)

    try:
        from .jmetal.ea import EA as JMetalEA
        engines['jmetal'] = JMetalEA
    except ImportError as e:
        print("jmetal not available")
        print(e)


algorithms = {'inspyred': ['SA', 'GA', 'NSGAII'],
              'jmetal': ['SA', 'GA', 'NSGAII', 'SPEA2', 'NSGAIII']
              }


def get_default_engine():
    global default_engine
    global engines

    if default_engine:
        return default_engine

    engine_order = ['jmetal', 'inspyred']

    for engine in engine_order:
        if engine in list(engines.keys()):
            default_engine = engine
            break

    if not default_engine:
        raise RuntimeError("No EA engine available.")

    return default_engine


def set_default_engine(enginename):
    """ Sets default EA engine.

    :param str enginename: Optimization engine (currently available: 'inspyred', 'jmetal')
    """

    global default_engine
    global engines

    if enginename.lower() in list(engines.keys()):
        default_engine = enginename.lower()
    else:
        raise RuntimeError(f"EA engine {enginename} not available.")


def set_preferred_EA(algorithm):
    """Defines de preferred MOEA.

    :param str algorithm: The name of the preferred algorithm.
    """
    global preferred_EA
    global default_engine
    global engines

    if algorithm in algorithms[get_default_engine()]:
        preferred_EA = algorithm
    else:
        for eng in engines.keys():
            if algorithm in algorithms[eng]:
                preferred_EA = algorithm
                default_engine = eng
                return
        raise ValueError(f"Algorithm {algorithm} is unavailable.")


def get_preferred_EA():
    """
    :returns: The name of the preferred MOEA.
    """
    global preferred_EA
    return preferred_EA


def get_available_engines():
    """
    :returns: The list of available engines.
    """
    global engines
    return list(engines.keys())


def get_available_algorithms():
    """
    :returns: The list of available MOEAs.
    """
    global engines
    algs = []
    for engine in engines.keys():
        algs.extend(algorithms[engine])
    return list(set(algs))


def EA(problem, initial_population=[], max_generations=EAConstants.MAX_GENERATIONS, mp=True, visualizer=False,
       algorithm=None, **kwargs):
    """
    EA running helper. Returns an instance of the EA that reflects the global user configuration settings
    such as preferred engine and algorithm.

    :param problem: The optimization problem.
    :param list initial_population: The EA initial population. Default [].
    :param int max_generations: The number of iterations of the EA (stopping criteria). Default globally defined.
    :param bool mp: If multiprocessing should be used. Default True.
    :param bool visualizer: If the pareto font should be displayed. Requires a graphic environment. Default False.

    Additional optional arguments:

    :param int population_size: EA population size.

    :returns: An instance of an EA optimizer.

    """
    global engines

    if len(engines) == 0:
        check_engines()
        if len(engines) == 0:
            raise RuntimeError('Inspyred or JMetal packages are required')

    if algorithm is None or algorithm not in get_available_algorithms():
        algorithm = get_preferred_EA()

    engs = [k for k, v in algorithms.items() if algorithm in v]

    if get_default_engine() in engs:
        engine = engines[get_default_engine()]
    else:
        engine = engines[engs[0]]

    return engine(problem, initial_population=initial_population, max_generations=max_generations, mp=mp,
                  visualizer=visualizer, algorithm=algorithm, **kwargs)


check_engines()
default_engine = None
preferred_EA = 'NSGAII'



import os
from collections import OrderedDict
from math import isinf, isnan

from libsbml import AssignmentRule, SBMLReader

from ..model.kinetic import ODEModel, Compartment, Metabolite, KineticReaction, Rule


def load_sbml(filename):
    """ Loads an SBML file.

    :param filename: SBML file path, str.
    :returns: SBMLModel

    """

    if not os.path.exists(filename):
        raise IOError("Model file was not found")

    reader = SBMLReader()
    document = reader.readSBML(str(filename))
    sbml_model = document.getModel()

    if sbml_model is None:
        document.printErrors()
        raise IOError(f'Failed to load model {filename}.')

    return sbml_model


def load_ODEModel(filename):
    sbml_model = load_sbml(filename)
    ode_model = ODEModel(sbml_model.getId())
    _load_compartments(sbml_model, ode_model)
    _load_metabolites(sbml_model, ode_model)
    # adds constants and boundaries to metabolites
    for species in sbml_model.getListOfSpecies():
        m = ode_model.metabolites[species.getId()]
        if not hasattr(m, "constant"):
            m.constant = species.getConstant()
        if not hasattr(m, "boundary"):
            m.boundary = species.getBoundaryCondition()
    # load_reactions(sbml_model, ode_model)
    _load_concentrations(sbml_model, ode_model)
    _load_global_parameters(sbml_model, ode_model)
    _load_ratelaws(sbml_model, ode_model)
    _load_assignment_rules(sbml_model, ode_model)
    return ode_model


def extract_metadata(sbml_elem, elem):

    sboterm = sbml_elem.getSBOTermID()
    if sboterm:
        elem.metadata['SBOTerm'] = sboterm

    notes = sbml_elem.getNotes()
    if notes:
        recursive_node_parser(notes, elem.metadata)

    annotation = sbml_elem.getAnnotationString()
    if annotation:
        elem.metadata['XMLAnnotation'] = annotation


def recursive_node_parser(node, cache):
    node_data = node.getCharacters()
    if ':' in node_data:
        key, value = node_data.split(':', 1)
        cache[key.strip()] = value.strip()

    for i in range(node.getNumChildren()):
        recursive_node_parser(node.getChild(i), cache)


def _load_compartments(sbml_model, model):
    for compartment in sbml_model.getListOfCompartments():
        model.add_compartment(_load_compartment(compartment))


def _load_compartment(compartment):
    size = compartment.getSize()
    if isnan(size) or isinf(size):
        size = 1.0

    comp = Compartment(compartment.getId(), compartment.getName(), False, size)
    extract_metadata(compartment, comp)
    return comp


def _load_metabolites(sbml_model, model):
    for species in sbml_model.getListOfSpecies():
        model.add_metabolite(_load_metabolite(species))


def _load_metabolite(species):
    metabolite = Metabolite(species.getId(), species.getName(), species.getCompartment())
    try:
        fbc_species = species.getPlugin('fbc')
        if fbc_species.isSetChemicalFormula():
            formula = fbc_species.getChemicalFormula()
            metabolite.metadata['FORMULA'] = formula

        if fbc_species.isSetCharge():
            charge = fbc_species.getCharge()
            metabolite.metadata['CHARGE'] = str(charge)
    except Exception:
        pass
    extract_metadata(species, metabolite)
    return metabolite


def _load_concentrations(sbml_model, odemodel):
    for species in sbml_model.getListOfSpecies():
        odemodel.set_concentration(species.getId(), species.getInitialConcentration())


def _load_global_parameters(sbml_model, odemodel):
    for parameter in sbml_model.getListOfParameters():
        odemodel.set_global_parameter(parameter.getId(), parameter.getValue(), parameter.getConstant())


def _load_ratelaws(sbml_model, odemodel):
    for reaction in sbml_model.getListOfReactions():
        formula = reaction.getKineticLaw().getFormula()
        parameters = OrderedDict()
        modifiers = []
        stoichiometry = OrderedDict()

        for reactant in reaction.getListOfReactants():
            m_id = reactant.getSpecies()
            coeff = -reactant.getStoichiometry()

            if m_id not in stoichiometry:
                stoichiometry[m_id] = coeff
            else:
                stoichiometry[m_id] += coeff

        for product in reaction.getListOfProducts():
            m_id = product.getSpecies()
            coeff = product.getStoichiometry()
            if m_id not in stoichiometry:
                stoichiometry[m_id] = coeff
            else:
                stoichiometry[m_id] += coeff
            if stoichiometry[m_id] == 0.0:
                del stoichiometry[m_id]

        for parameter in reaction.getKineticLaw().getListOfParameters():
            parameters[parameter.getId()] = parameter.getValue()
        for modifier in reaction.getListOfModifiers():
            m_id = modifier.getSpecies()
            modifiers.append(m_id)

        law = KineticReaction(reaction.getId(), formula, name=reaction.getName(), 
                              stoichiometry=stoichiometry,
                              parameters=parameters, modifiers=modifiers, 
                              reversible=reaction.getReversible())
        odemodel.set_ratelaw(reaction.getId(), law)


def _load_assignment_rules(sbml_model, odemodel):
    for rule in sbml_model.getListOfRules():
        if isinstance(rule, AssignmentRule):
            r_id = rule.getVariable()
            odemodel.set_assignment_rule(r_id, Rule(r_id, rule.getFormula()))



from typing import Union, Dict, Iterable, Tuple, TYPE_CHECKING, Sequence
from io import TextIOWrapper

from cobra import Model as Cobra_Model
from reframed import CBModel as Reframed_Model
from mewpy.io import Reader, Engines, read_model
from .problem import AbstractKOProblem


if TYPE_CHECKING:
    from mewpy.optimization import EvaluationFunction
    from mewpy.germ.models import Model, MetabolicModel, RegulatoryModel


def load_optorf(regulatory_model: Union[str, TextIOWrapper, Reader],
                metabolic_model: Union[str, TextIOWrapper, Cobra_Model, Reframed_Model, Reader],
                config: dict = None,
                warnings: bool = False):
    """
    The standard method to load an OptORF problem.
    A OptORF problem is a KO strain optimization problem
    that can be created from an integrated Metabolic-Regulatory model.

    The integrated model can be assembled from a metabolic model in the following formats:
        - SBML
        - JSON
        - Cobrapy model
        - Reframed model

    and a regulatory model in the following formats:
        - SBML-qual
        - JSON
        - CSV regulatory network

    :param regulatory_model: the regulatory model or a reader for parsing it
    :param metabolic_model: the metabolic model or a reader for parsing it
    :param config: A configuration dictionary with information for parsing the models
    :param warnings: if True, warnings will be printed
    :return: an integrated model for OptORF
    """
    if not config:
        config = {}

    if not isinstance(regulatory_model, Reader):

        if isinstance(regulatory_model, str):

            file_name = regulatory_model

        elif isinstance(regulatory_model, TextIOWrapper):

            file_name = regulatory_model.name

        else:
            raise ImportError('Invalid file type')

        engine = Engines.BooleanRegulatoryCSV

        if file_name.endswith('.xml') or file_name.endswith('.sbml'):
            engine = Engines.RegulatorySBML

        regulatory_model = Reader(engine=engine,
                                  io=regulatory_model,
                                  **config)

    if not isinstance(metabolic_model, Reader):

        if isinstance(metabolic_model, str):

            engine = Engines.MetabolicSBML

        elif isinstance(metabolic_model, TextIOWrapper):

            engine = Engines.MetabolicSBML

        elif isinstance(metabolic_model, Cobra_Model):

            engine = Engines.CobraModel

        elif isinstance(metabolic_model, Reframed_Model):

            engine = Engines.ReframedModel

        else:
            raise ImportError('Invalid file type')

        metabolic_model = Reader(engine=engine,
                                 io=metabolic_model,
                                 **config)

    return read_model(regulatory_model, metabolic_model, warnings=warnings)


class OptORFProblem(AbstractKOProblem):

    def __init__(self,
                 model: Union['Model', 'MetabolicModel', 'RegulatoryModel'],
                 fevaluation: Sequence['EvaluationFunction'],
                 initial_state: Dict[str, float] = None,
                 **kwargs):

        """
        OptORF problem using the RFBA implementation analysis.
        The OptORF approach is based on gene and regulator deletion to identify optimization strategies.

        It currently supports KOProblem-based decoding

        For more details consult: https://doi.org/10.1186/1752-0509-4-53
        """
        if isinstance(model, (Cobra_Model, Reframed_Model)):
            raise ValueError(f'OptORF is not available for a model of type {type(model)}.'
                             f'Please use load_optorf() to retrieve an integrated GERM model')

        super(OptORFProblem, self).__init__(model, fevaluation, **kwargs)

        # ignore user defined target list
        self._trg_list = None

        if not initial_state:
            initial_state = {}
        self._initial_state = initial_state.copy()

    def _build_target_list(self):
        """
        Build the target list and initial state for the OptORF problem.
        The target list is the list of regulators in the regulatory layer.
        The initial state is the state of the model before any gene deletion.
        :return:
        """
        # Target list is the combination of genes and regulators available into the mewpy integrated model

        regulators = [regulator.id for regulator in self.model.yield_regulators()
                      if not regulator.is_reaction() and not regulator.is_metabolite()]
        targets = [target.id for target in self.model.yield_targets()
                   if not target.is_reaction() and not target.is_metabolite()]
        genes = [gene.id for gene in self.model.yield_genes()
                 if not gene.is_reaction() and not gene.is_metabolite()]

        self._trg_list = list(set.union(set(regulators), set(targets), set(genes)))

    def _decode_initial_state(self, state: Dict[str, float] = None) -> Dict[str, float]:
        """
        Method responsible for retrieving the initial state of the model.
        The initial state is the state of all regulators found in the Metabolic-Regulatory model.
        :param state: the initial state of the model
        :return: dict of regulatory/metabolic variable keys (regulators) and a value of 0 or 1
        """
        if not state:
            state = {}

        if not self.model.is_regulatory():
            return state

        initial_state = {}
        for regulator in self.model.yield_regulators():
            if regulator.id in state:
                initial_state[regulator.id] = state[regulator.id]

            elif regulator.is_metabolite() and regulator.exchange_reaction:
                if regulator.exchange_reaction.id in state:
                    initial_state[regulator.id] = state[regulator.exchange_reaction.id]

                else:
                    initial_state[regulator.id] = abs(regulator.exchange_reaction.lower_bound)

            else:
                initial_state[regulator.id] = max(regulator.coefficients)

        return initial_state

    def _decode_regulatory_state(self, state: Dict[str, float]) -> Dict[str, float]:
        """
        It solves the boolean regulatory network for a given specific state.
        It also updates all targets having a valid regulatory interaction associated with it for the resulting state

        :param state: dict of regulatory variable keys (regulators) and a value of 0, 1 or float
        (reactions and metabolites predicates)
        :return: dict of target keys and a value of the resulting state
        """
        if not self.model.is_regulatory():
            return {}

        # Solving regulatory model synchronously for the regulators according to the initial state
        # Targets are associated with a single regulatory interaction
        result = {}
        for interaction in self.model.yield_interactions():

            # solving regulators state only
            if not interaction.target.is_regulator():
                continue

            # an interaction can have multiple regulatory events, namely one for 0 and another for 1
            for coefficient, event in interaction.regulatory_events.items():
                if event.is_none:
                    continue

                eval_result = event.evaluate(values=state)
                if eval_result:
                    result[interaction.target.id] = coefficient
                else:
                    result[interaction.target.id] = 0.0
        return result

    def _decode_metabolic_state(self, state: Dict[str, float]) -> Dict[str, float]:
        """
        It solves the boolean regulatory network for a given specific state.
        It also updates all targets having a valid regulatory interaction associated with it for the resulting state

        :param state: dict of regulatory variable keys (regulators) and a value of 0, 1 or float
        (reactions and metabolites predicates)
        :return: dict of target keys and a value of the resulting state
        """
        if not self.model.is_regulatory():
            return {}

        # Solving the whole regulatory model synchronously, as asynchronously would take too much time
        # Targets are associated with a single regulatory interaction
        result = {}
        for interaction in self.model.yield_interactions():

            # an interaction can have multiple regulatory events, namely one for 0 and another for 1
            for coefficient, event in interaction.regulatory_events.items():
                if event.is_none:
                    continue

                eval_result = event.evaluate(values=state)
                if eval_result:
                    result[interaction.target.id] = coefficient
                else:
                    result[interaction.target.id] = 0.0
        return result

    def _decode_constraints(self, state: Dict[str, float]) -> Dict[str, Tuple[float, float]]:
        """
        Method responsible for decoding the RFBA metabolic state, namely the state of all metabolic genes associated
        at least with one reaction in the GPRs rule.

        :param state: dict of regulatory/metabolic variable keys (metabolic target) and a value of 0 or 1
        :return: dict of constraints of the resulting metabolic state
        """
        # This method retrieves the constraints associated with a given metabolic/regulatory state

        if not self.model.is_metabolic():
            return {}

        constraints = {}
        for rxn in self.model.yield_reactions():

            if rxn.gpr.is_none:
                continue

            res = rxn.gpr.evaluate(values=state)

            if not res:
                constraints[rxn.id] = (0.0, 0.0)

        return constraints

    def decode(self, candidate: Iterable[int]) -> Dict[str, Tuple[float, float]]:
        """
        OptORF candidate decode.
        From a candidate aka a set of regulators to be knocked out, it runs a synchronous boolean simulation of the
        regulatory layer. The regulatory state is used to infer the metabolic state, namely metabolic reactions affected
        by the KOs. A second synchronous boolean simulation of the GPRs is performed with the regulatory state.
        Finally, the metabolic state aka constraints are decoded from the inactive reactions.

        :param candidate: a candidate (a set of regulators to be knocked out)
        :return: a metabolic state (a set of constraints)
        """
        # the perturbation state to be used to decode the initial state for simulation
        candidate_state = {self.target_list[idx]: 0 for idx in candidate}

        # the initial state of the model
        initial_state = {**self._initial_state, **candidate_state}
        initial_state = self._decode_initial_state(state=initial_state)

        # Regulatory state from a synchronous boolean simulation
        # noinspection PyTypeChecker
        regulatory_state = self._decode_regulatory_state(state=initial_state)

        # After a simulation of the regulators outputs, the state of the targets are retrieved now
        metabolic_state = self._decode_metabolic_state(state={**initial_state, **regulatory_state})

        regulatory_constraints = self._decode_constraints(metabolic_state)
        return regulatory_constraints



import joblib
import contextlib
import functools
import re
import types
import time
from collections.abc import Iterable
from .constants import atomic_weights
from warnings import warn

class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

    def find(self, pattern=None, sort=False):
        """A user friendly method to find metabolites, reactions or genes in the model.

        :param pattern: The pattern which can be a regular expression, defaults to None in which case all entries are listed.
        :type pattern: str, optional
        :param sort: if the search results should be sorted, defaults to False
        :type sort: bool, optional
        :return: the search results
        :rtype: pandas dataframe
        """
        values = list(self.keys())
        if pattern:
            import re
            if isinstance(pattern, list):
                patt = '|'.join(pattern)
                re_expr = re.compile(patt)
            else:
                re_expr = re.compile(pattern)
            values = [x for x in values if re_expr.search(x) is not None]
        if sort:
            values.sort()

        import pandas as pd
        data = [{'attribute':x,'value':self.get(x)} for x in values]
        
        if data:
            df = pd.DataFrame(data)
            df = df.set_index(df.columns[0])
        else: 
            df = pd.DataFrame()
        return df

    def __repr__(self) -> str:
        return str(self.find())


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""


class Timer:
    def __init__(self):
        self._start_time = None

    def start(self):
        """Start a new timer"""
        if self._start_time is not None:
            raise TimerError("Timer is running. Use .stop() to stop it")

        self._start_time = time.perf_counter()

    def stop(self):
        """Stop the timer, and report the elapsed time"""
        if self._start_time is None:
            raise TimerError("Timer is not running. Use .start() to start it")

        elapsed_time = time.perf_counter() - self._start_time
        self._start_time = None
        print(f"Elapsed time: {elapsed_time:0.6f} seconds")

    def __enter__(self):
        """Start a new timer as a context manager"""
        self.start()
        return self

    def __exit__(self, *exc_info):
        """Stop the context manager timer"""
        self.stop()


class Singleton(object):
    _instance = None

    def __new__(class_, *args, **kwargs):
        if not isinstance(class_._instance, class_):
            class_._instance = object.__new__(class_, *args, **kwargs)
        return class_._instance


def copy_func(f):
    """Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)"""
    g = types.FunctionType(f.__code__, f.__globals__, name=f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__)
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return g


class Dispatcher:

    def __init__(self):
        """
        Dispatcher for the simulate method of the Simulation interface
        It allows a simplification of the if else chain of methods provided as input to the simulate method

        based on https://stackoverflow.com/questions/36836161/singledispatch-based-on-value-instead-of-type
        (Ilja Everilä)

        """

        # weak ref key for the garbage collector
        self.registry = {}

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self.dispatch(instance, owner)

    def dispatch(self, instance, owner):
        def wrapper(state, *args, **kwargs):
            method = self.registry.get(state).__get__(instance, owner)
            return method(*args, **kwargs)
        return wrapper

    def register(self, state):
        def wrapper(method):
            self.registry[state] = method
            return method
        return wrapper


def iterable(obj, is_string=False):
    if isinstance(obj, Iterable):
        if is_string and isinstance(obj, str):
            return (obj,)
        return obj
    return (obj,)

def generator(container):
    return (value for value in container.values())

# Taken from the talented team responsible for developing cobrapy!!!!
chemical_formula_re = re.compile('([A-Z][a-z]?)([0-9.]+[0-9.]?|(?=[A-Z])?)')

def elements(formula):
    all_elements = re.findall(chemical_formula_re, formula)
    atoms = {}
    for atom, count in all_elements:
        if not count:
            count = '1'
        atoms[atom] = atoms.get(atom, 0) + int(count)
    return atoms

def molecular_weight(formula, element=None):
    elems = elements(formula)
    if element:
        mw = elems.get(element, 0) * atomic_weights.get(element, 0)
    else:
        missing = set(elems) - set(atomic_weights)
        if missing:
            warn(f"Atomic weight not listed for elements: {missing}")

        mw = sum(atomic_weights.get(elem, 0) * n for elem, n in elems.items())

    return mw

@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""
    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


def get_all_subclasses(cls):
    '''Returns all subclasses of a class'''
    all_subclasses = []

    for subclass in cls.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(get_all_subclasses(subclass))

    return all_subclasses

def is_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter



from mewpy.problems.problem import AbstractKOProblem, AbstractOUProblem
from mewpy.optimization.evaluation import KineticEvaluationFunction
from mewpy.simulation.kinetic import KineticSimulation
from mewpy.solvers import KineticConfigurations
from re import search

class KineticKOProblem(AbstractKOProblem):

    def __init__(self, model, fevaluation=None, **kwargs):
        super(KineticKOProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)
        kinetic_parameters = kwargs.get('kparam', None)
        t_points = kwargs.get('t_points', None)
        timeout = kwargs.get('timeout', KineticConfigurations.SOLVER_TIMEOUT)
        self.kinetic_sim = KineticSimulation(model, 
                                             parameters=kinetic_parameters,
                                             timeout=timeout,
                                             t_points=t_points)
        for f in self.fevaluation:
            if isinstance(f,KineticEvaluationFunction):
                f.kinetic = True
            else:
                raise ValueError(f"The optimization function {f} is not intended for kinetic optimization.")

    def _build_target_list(self):
        p = list(self.model.get_parameters(exclude_compartments=True))
        target =[]
        for k in p:
            if search(r'(?i)[rv]max',k):
                target.append(k)
        if self.non_target is not None:
            target = target - set(self.non_target)
        self._trg_list = list(target)

    def decode(self, candidate):
        factors = {self.target_list[idx]: 0 for idx in candidate}
        return factors

    def evaluate_solution(self, solution, decode=True):
        p = []
        factors = self.decode(solution) if decode else solution
        simulation_results = self.kinetic_sim.simulate(factors=factors)
        for f in self.fevaluation:
            p.append(f(simulation_results, factors))
        return p


class KineticOUProblem(AbstractOUProblem):

    def __init__(self, model, fevaluation=None, **kwargs):
        super(KineticOUProblem, self).__init__(
            model, fevaluation=fevaluation, **kwargs)
        kinetic_parameters = kwargs.get('kparam', None)
        t_points = kwargs.get('t_points', None)
        timeout = kwargs.get('timeout', KineticConfigurations.SOLVER_TIMEOUT)
        self.kinetic_sim = KineticSimulation(model, 
                                             parameters=kinetic_parameters,
                                             timeout=timeout,
                                             t_points=t_points)
        for f in self.fevaluation:
            if isinstance(f,KineticEvaluationFunction):
                f.kinetic = True
            else:
                raise ValueError(f"The optimization function {f} is not intended for kinetic optimization.")

    def _build_target_list(self):
        p = list(self.model.get_parameters(exclude_compartments=True))
        target =[]
        for k in p:
            if search(r'(?i)max',k):
                target.append(k)
        if self.non_target is not None:
            target = set(target) - set(self.non_target)
        self._trg_list = list(target)

    def decode(self, candidate):
        factors = {self.target_list[idx]: self.levels[lv_idx] for idx, lv_idx in candidate}
        return factors

    def evaluate_solution(self, solution, decode=True):
        p = []
        factors = self.decode(solution) if decode else solution
        simulation_results = self.kinetic_sim.simulate(factors=factors)
        for f in self.fevaluation:
            p.append(f(simulation_results, factors))
        return p



"""
##############################################################################
File containing string exoression parsing utility functions.
Expressions are decomposed into parsing trees for infix evaluation.

Authors: Vitor Pereira
##############################################################################
"""
import re
import sys
from abc import abstractmethod
from operator import add, sub, mul, truediv, pow

# Boolean operator symbols
S_AND = '&'
S_OR = '|'
S_NOT = '~'
S_ON = '1'
S_OFF = '0'
S_GREATER = '>'
S_LESS = '<'
S_EQUAL = '='
S_GREATER_THAN_EQUAL = '>='
S_LESS_THAN_EQUAL = '=>'

# Empty leaf symbol
EMPTY_LEAF = '@'

BOOLEAN_OPERATORS = [S_AND, S_OR, S_NOT]

BOOLEAN_STATES = [S_ON, S_OFF]

BOOLEAN_RELATIONALS = [S_GREATER, S_LESS, S_EQUAL]
BOOLEAN_EQUAL_RELATIONALS = [S_GREATER_THAN_EQUAL, S_LESS_THAN_EQUAL]

# Increases the system recursion limit for long expressions
sys.setrecursionlimit(100000)


def evaluate_expression(expression, variables):
    """ Evaluates a logical expression (containing variables, 'and','or','(' and ')')
    against the presence (True) or absence (False) of propositions within a list.
    The evaluation is achieved usind python native eval function.

    :param str expression: The expression to be evaluated.
    :param list variables: List of variables to be evaluated as True.
    :returns: A boolean evaluation of the expression.

    """
    expression = expression.replace('(', '( ').replace(')', ' )')
    # Symbol conversion not mandatory
    expression = expression.replace('and', '&').replace('or', '|')
    tokens = expression.split()
    sentence = []
    for token in tokens:
        if token in "&|()":
            sentence.append(token)
        elif token in variables:
            sentence.append("True")
        else:
            sentence.append("False")
    proposition = ' '.join(sentence)
    return eval(proposition)


def evaluate_expression_tree(expression, variables):
    """ Evaluates a logical expression (containing variables, 'and','or', 'not','(' , ')') against the presence (True)
    or absence (False) of propositions within a list.
    Assumes the correctness of the expression. The evaluation is achieved by means of a parsing tree.

    :param str expression: The expression to be evaluated.
    :param list variables: List of variables to be evaluated as True.
    :returns: A boolean evaluation of the expression.

    """
    t = build_tree(expression, Boolean)
    evaluator = BooleanEvaluator(variables)
    res = t.evaluate(evaluator.f_operand, evaluator.f_operator)
    return res


def maybe_fn(f, v1, v2):
    """Maybe evaluator: if one of the arguments is None, it retuns the value of the other\
    argument. If both arguments are None, it returns None. If both arguments are not None\
    it returns the evaluation f(v1,v2).

    :param f: a function
    :param v1: the first argument
    :param v2: the second argument  
    """
    if v1 is None:
        return v2
    elif v2 is None:
        return v1
    else:
        return f(v1, v2)


class Node(object):
    """
    Binary syntax tree node.

    :param value: The node value.
    :param left: The left node or None.
    :param right: The right node or None.
    """

    def __init__(self, value, left=None, right=None, tp=0):
        """Binary tree.  Empty leafs with None or 'EMPTY_LEAF' are always left sided.

        Args:
            value : Node value
            left (Node, optional): [description]. Defaults to None.
            right (Node, optional): [description]. Defaults to None.
            tp (int, optional): 1-Unary function, 2-binary function. Defaults to 0.

        Raises:
            ValueError
        """
        if left is not None and not isinstance(left, Node):
            raise ValueError("Invalid right element")
        if right is not None and not isinstance(right, Node):
            raise ValueError("Invalid right element")
        self.value = value
        self.left = left
        self.right = right
        self.tp = tp

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if self.is_leaf():
            return str(self.value)
        else:
            return f"{str(self.value)} ( {str(self.left)} , {str(self.right)} )"

    def is_leaf(self):
        """
        :returns: True if the node is a leaf False otherwise. Both left and right are None.
        """
        return not self.left and not self.right

    def is_empty_leaf(self):
        """
        :returns: True if the node is an empty leaf False otherwise.
        """
        return self.value == EMPTY_LEAF

    def is_unary(self):
        """
        :returns: True if the node is a unary operation False otherwise.
        """
        return (self.left.is_empty_leaf() and not self.right.is_empty_leaf()) or (
            not self.left.is_empty_leaf() and self.right.is_empty_leaf())

    def is_binary(self):
        """
        :returns: True if the node is a binary operation False otherwise.
        """
        return not self.left.is_empty_leaf() and not self.right.is_empty_leaf()

    def get_operands(self):
        if self.is_leaf():
            if self.value == EMPTY_LEAF:
                return set()
            else:
                return {self.value}
        else:
            return self.left.get_operands().union(self.right.get_operands())

    def get_parameters(self):
        """Parameters are all non numeric symbols in an expression"""
        if self.is_leaf():
            if self.value != EMPTY_LEAF and not is_number(self.value):
                return {self.value}
            else:
                return set()
        else:
            return self.left.get_parameters().union(self.right.get_parameters())

    def print_node(self, level=0):
        """Prints a parsing tree of the expression."""
        tabs = ""
        for _ in range(level):
            tabs += "\t"
        if self.is_leaf():
            if self.value != EMPTY_LEAF:
                print(tabs, f"|____{self.value}")
            else:
                pass
        else:
            print(tabs, f"|____{self.value}")
        if self.left is not None:
            self.left.print_node(level + 1)
        if self.right is not None:
            self.right.print_node(level + 1)

    def evaluate(self, f_operand=None, f_operator=None):
        """
        Evaluates the expression using the f_operand and f_operator mapping functions
        """
        if f_operand is None or f_operator is None:
            return eval(str(self))
        elif self.is_leaf():
            return f_operand(self.value)
        else:
            return maybe_fn(f_operator(self.value),
                            self.left.evaluate(f_operand, f_operator),
                            self.right.evaluate(f_operand, f_operator)
                            )

    def get_conditions(self):
        """
        Retrieves the propositional conditions

        return: The list of conditions
        """
        ops = self.get_operands()
        return set([i for i in ops if is_condition(i)])

    def replace(self, r_map: dict):
        """Apply the mapping replacing to the tree

        Args:
            map (dict): replacement mapping

        Returns:
            Node: new Tree with replace entries
        """
        if self.is_leaf():
            v = r_map[self.value] if self.value in r_map.keys() else self.value
            return Node(v, None, None)
        else:
            return Node(self.value, self.left.replace(r_map), self.right.replace(r_map), self.tp)

    def to_infix(self) -> str:
        if self.is_leaf():
            if self.value == EMPTY_LEAF:
                return ''
            else:
                return str(self.value)
        elif self.tp == 2:
            return ''.join([self.value, '( ', self.left.to_infix(), ' , ', self.right.to_infix(), ' )'])
        elif self.tp == 1:
            return ''.join([self.value, '( ', self.right.to_infix(), ' )'])
        else:
            return ''.join(['( ', self.left.to_infix(), ' ', self.value, ' ', self.right.to_infix(), ' )'])

    def copy(self):
        if self.is_leaf():
            return Node(self.value.copy(), None, None)
        else:
            return Node(self.value.copy(), self.left.copy(), self.right.copy(), self.tp)


class Syntax:
    """Defines an interface for the tree syntax parsing
       with operators, their precedence and associativity.
    """

    @ abstractmethod
    def is_operator(self, op):
        raise NotImplementedError

    @ abstractmethod
    def is_greater_precedence(self, op1, op2):
        raise NotImplementedError

    @ abstractmethod
    def associativity(self, op):
        raise NotImplementedError

    @ staticmethod
    def arity(op):
        return 2

    @ staticmethod
    def replace():
        return {}


class Arithmetic(Syntax):
    """Defines a basic arithmetic sintax.
    """

    @ staticmethod
    def is_operator(op):
        return op in ['+', '-', '*', '/', '^']

    @ staticmethod
    def is_greater_precedence(op1, op2):
        pre = {'+': 0, '-': 0, '*': 1, '/': 1, '^': 2}
        return pre[op1] >= pre[op2]

    @ staticmethod
    def associativity(op):
        ass = {'+': 0, '-': 0, '*': 0, '/': 0, '^': 1}
        return ass[op]

    @ staticmethod
    def arity(op):
        ar = {'+': 2, '-': 2, '*': 2, '/': 2, '^': 2}
        return ar[op]


class ArithmeticEvaluator:

    @ staticmethod
    def f_operator(op):
        operators = {'+': add, '-': sub, '*': mul, '/': truediv, '^': pow, 'pow': pow}
        if op in operators.keys():
            return operators[op]
        else:
            raise ValueError(f"Operator {op} not defined")

    @ staticmethod
    def f_operand(op):
        if is_number(op):
            return float(op)
        else:
            raise ValueError(f"Operator {op} not defined")


class Boolean(Syntax):
    """
    A boolean syntax parser where (NOT) is considered to
    be defined as a binary operator (NOT a) = (EMPTY NOT a)
    """

    @ staticmethod
    def is_operator(op):
        return op in [S_AND, S_OR, S_NOT]

    @ staticmethod
    def is_greater_precedence(op1, op2):
        pre = {S_AND: 0, S_OR: 0, S_NOT: 1}
        return pre[op1] >= pre[op2]

    @ staticmethod
    def associativity(op):
        ass = {S_AND: 0, S_OR: 0, S_NOT: 1}
        return ass[op]

    @ staticmethod
    def arity(op):
        ar = {S_AND: 2, S_OR: 2, S_NOT: 1}
        return ar[op]

    @ staticmethod
    def replace():
        r = {'not': [EMPTY_LEAF, S_NOT], 'and': [S_AND],
             'or': [S_OR], '&': [S_AND], '|': [S_OR], '~': [EMPTY_LEAF, S_NOT]}
        return r

class BooleanEvaluator:
    """A boolean evaluator.

    :param list true_list: Operands evaluated as True. Operands not in the list are evaluated as False
    :param dict variables: A dictionary mapping symbols to values. Used to evaluate conditions.

    """

    def __init__(self, true_list=[], variables={}):
        self.true_list = true_list
        self.vars = variables

    def f_operator(self, op):
        operators = {S_AND: lambda x, y: x and y, S_OR: lambda x, y: x or y, S_NOT: lambda x, y: not y}
        if op in operators.keys():
            return operators[op]
        else:
            raise ValueError(f"Operator {op} not defined")

    def f_operand(self, op):
        if op.upper() == 'TRUE' or op == '1' or op in self.true_list:
            return True
        elif is_condition(op):
            return eval(op, None, self.vars)
        else:
            return False

    def set_true_list(self, true_list):
        self.true_list = true_list


class GeneEvaluator:
    """An evaluator for genes expression.

    :param genes_value: A dictionary mapping genes to values. Gene not in the dictionary have a value of 1.
    :param and_operator: function to be applied instead of (and', '&') (not case sensitive)
    :param or_operator: function to be applied instead of ('or','|') (not case sensitive)
    """

    def __init__(self, genes_value, and_operator=min, or_operator=max, prefix="", unexpressed_value=1):

        self.genes_value = genes_value
        self.and_operator = and_operator
        self.or_operator = or_operator
        self.prefix = prefix
        self.unexpressed_value = unexpressed_value

    def f_operator(self, op):
        operators = {S_AND: self.and_operator, S_OR: self.or_operator}
        if op in operators.keys():
            return operators[op]
        else:
            raise ValueError(f"Operator {op} not defined")

    def f_operand(self, op):
        if op[len(self.prefix):] in self.genes_value:
            return self.genes_value[op]
        else:
            return self.unexpressed_value


def tokenize_function(exp: str):
    p = 0
    s = -1
    tokens = []
    i = 0
    while i < len(exp):
        if exp[i] == '(':
            s += 1
            if s == 0:
                p = i+1
            if not tokens:
                tokens = [exp[:i]]
        elif exp[i] == ')':
            s -= 1
            if s == -1:
                tokens.append(exp[p:i])
        elif exp[i] == ',' and s == 0:
            tokens.append(exp[p:i])
            p = i+1
        i += 1
    if not tokens:
        return [exp]
    else:
        return tokens


# Tree
def build_tree(exp, rules):
    """ Builds a parsing syntax tree for basic mathematical expressions

    :param exp: the expression to be parsed
    :param rules: Sintax definition rules
    """
    replace_dic = rules.replace()
    exp_ = tokenize_infix_expression(exp)
    exp_list = []
    for token in exp_:
        if token.lower() in replace_dic:
            exp_list.extend(replace_dic[token.lower()])
        else:
            exp_list.append(token)
    stack = []
    tree_stack = []
    predecessor = None
    i = 0
    while i < len(exp_list):
        token = exp_list[i]
        if not (rules.is_operator(token) or token in ['(', ')']):
            if i < len(exp_list)-2 and exp_list[i+1] == '(':
                s = 1
                p = i+2
                while p < len(exp_list) and s > 0:
                    if exp_list[p] == '(':
                        s += 1
                    elif exp_list[p] == ')':
                        s -= 1
                    p += 1
                token = ' '.join(exp_list[i:p])
                i = p - 1

            if predecessor and not (rules.is_operator(predecessor) or predecessor in ['(', ')']):
                s = tree_stack[-1].value
                tree_stack[-1].value = s + " " + token
            else:
                if '(' in token:
                    f = tokenize_function(token)
                    if len(f) == 2:
                        t = Node(f[0], Node(EMPTY_LEAF), build_tree(f[1], rules), 1)
                    elif len(f) == 3:
                        t = Node(f[0], build_tree(f[1], rules), build_tree(f[2], rules), 2)
                    else:
                        t = Node(token)
                else:
                    t = Node(token)
                tree_stack.append(t)
        elif rules.is_operator(token):
            if not stack or stack[-1] == '(':
                stack.append(token)

            elif rules.is_greater_precedence(token, stack[-1]) and rules.associativity(token) == 1:
                stack.append(token)

            else:
                while stack and stack[-1] != '(' and rules.is_greater_precedence(stack[-1], token) \
                        and rules.associativity(token) == 0:
                    popped_item = stack.pop()
                    t = Node(popped_item)
                    t1 = tree_stack.pop()
                    t2 = tree_stack.pop()
                    t.right = t1
                    t.left = t2
                    tree_stack.append(t)
                stack.append(token)

        elif token == '(':
            stack.append('(')

        elif token == ')':
            while stack[-1] != '(':
                popped_item = stack.pop()
                t = Node(popped_item)
                t1 = tree_stack.pop()
                t2 = tree_stack.pop()
                t.right = t1
                t.left = t2
                tree_stack.append(t)
            stack.pop()
        predecessor = token
        i += 1

    while stack:
        popped_item = stack.pop()
        t = Node(popped_item)
        t1 = tree_stack.pop()
        t2 = tree_stack.pop()
        t.right = t1
        t.left = t2
        tree_stack.append(t)

    t = tree_stack.pop()

    return t


def tokenize_infix_expression(exp):
    return list(filter(lambda x: x != '', exp.replace('(', ' ( ').replace(')', ' ) ').split(' ')))


def is_number(token):
    """ Returns True if the token is a number
    """
    return token.replace('.', '', 1).replace('-', '', 1).isnumeric()


def is_condition(token):
    """ Returns True if the token is a condition
    """
    regexp = re.compile(r'>|<|=')
    return bool(regexp.search(token))



from mewpy.util.utilities import Singleton


class MEWPYSimSolvers(Singleton):

    def __init__(self):
        self._mewpy_sim_solvers = []

    def build(self):
        try:
            import gurobipy
            self._mewpy_sim_solvers.append('gurobi')
        except ImportError as e:
            pass
        try:
            import cplex
            self._mewpy_sim_solvers.append('cplex')
        except ImportError as e:
            pass

        try:
            import swiglpk
            self._mewpy_sim_solvers.append('glpk')
        except ImportError:
            pass

    def get_solvers(self):
        if not self._mewpy_sim_solvers:
            self.build()
        return self._mewpy_sim_solvers


__MEWPY_sim_solvers__ = MEWPYSimSolvers().get_solvers()



""" JMetal Problems
"""
import random
from typing import Tuple, List
from jmetal.core.problem import Problem
from jmetal.core.solution import Solution

from ..ea import SolutionInterface, dominance_test
from ...util.process import Evaluable


# define EA representation for OU
IntTupple = Tuple[int]


class KOSolution(Solution[int], SolutionInterface):
    """ Class representing a KO solution """

    def __init__(self, lower_bound: int, upper_bound: int, number_of_variables: int, number_of_objectives: int,
                 number_of_constraints: int = 0):
        super(KOSolution, self).__init__(number_of_variables,
                                         number_of_objectives, number_of_constraints)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def __eq__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return self.variables.sort() == solution.variables.sort()
        return False

    # JMetal consideres all problems as minimization
    # Based on pareto dominance

    def __gt__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) == 1
        return False

    def __lt__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) == -1
        return False

    def __ge__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) != -1
        return False

    def __le__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) != 1
        return False

    def __copy__(self):
        new_solution = KOSolution(
            self.lower_bound,
            self.upper_bound,
            self.number_of_variables,
            self.number_of_objectives)
        new_solution.objectives = self.objectives[:]
        new_solution.variables = self.variables[:]
        new_solution.constraints = self.constraints[:]
        new_solution.attributes = self.attributes.copy()

        return new_solution

    def get_representation(self):
        """
        Returns a set representation of the candidate
        """
        return set(self.variables)

    def get_fitness(self):
        """
        Returns the candidate fitness list
        """
        return self.objectives

    def __str__(self):
        return " ".join((self.variables))


class OUSolution(Solution[IntTupple], SolutionInterface):
    """
    Class representing a Over/Under expression solution.
    """

    def __init__(self, lower_bound: List[int], upper_bound: List[int], number_of_variables: int,
                 number_of_objectives: int):
        super(OUSolution, self).__init__(number_of_variables,
                                         number_of_objectives)
        self.upper_bound = upper_bound
        self.lower_bound = lower_bound

    def __eq__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return self.variables.sort() == solution.variables.sort()
        return False

    # JMetal consideres all problems as minimization

    def __gt__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) == 1
        return False

    def __lt__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) == -1
        return False

    def __ge__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) != -1
        return False

    def __le__(self, solution) -> bool:
        if isinstance(solution, self.__class__):
            return dominance_test(self, solution, maximize=False) != 1
        return False

    def __copy__(self):
        new_solution = OUSolution(
            self.lower_bound,
            self.upper_bound,
            self.number_of_variables,
            self.number_of_objectives
        )
        new_solution.objectives = self.objectives[:]
        new_solution.variables = self.variables[:]
        new_solution.constraints = self.constraints[:]
        new_solution.attributes = self.attributes.copy()

        return new_solution

    def get_fitness(self):
        """
        Returns the candidate fitness list
        """
        return self.objectives

    def __str__(self):
        return " ".join((self.variables))


class JMetalKOProblem(Problem[KOSolution], Evaluable):

    def __init__(self, problem, initial_polulation):
        """JMetal OU problem. Encapsulates a MEWpy problem
        so that it can be used in jMetal.
        """
        self.problem = problem
        self.number_of_objectives = len(self.problem.fevaluation)
        self.obj_directions = []
        self.obj_labels = []
        for f in self.problem.fevaluation:
            self.obj_labels.append(str(f))
            if f.maximize:
                self.obj_directions.append(self.MAXIMIZE)
            else:
                self.obj_directions.append(self.MINIMIZE)
        self.initial_polulation = initial_polulation
        self.__next_ini_sol = 0

    def create_solution(self) -> KOSolution:
        solution = None
        flag = False
        while self.__next_ini_sol < len(self.initial_polulation) and not flag:
            s = self.initial_polulation[self.__next_ini_sol]
            try:
                solution = self.problem.encode(s)
                flag = True
                self.__next_ini_sol += 1
            except ValueError as e:
                print("Skipping seed:", s, " ", e)
                self.__next_ini_sol += 1
        if not solution:
            solution = self.problem.generator(random)
        new_solution = KOSolution(
            self.problem.bounder.lower_bound,
            self.problem.bounder.upper_bound,
            len(solution),
            self.problem.number_of_objectives)
        new_solution.variables = list(solution)[:]
        return new_solution

    def reset_initial_population_counter(self):
        """ Resets the pointer to the next initial population element.
        This strategy is used to overcome the unavailable seeding API in jMetal.
        """
        import random
        random.shuffle(self.initial_polulation)
        self.__next_ini_sol = 0

    def get_constraints(self, solution):
        return self.problem.decode(set(solution.variables))

    def evaluate(self, solution: KOSolution) -> KOSolution:
        candidate = set(solution.variables)
        p = self.problem.evaluate_solution(candidate)
        for i in range(len(p)):
            # JMetalPy only deals with minimization problems
            if self.obj_directions[i] == self.MAXIMIZE:
                solution.objectives[i] = -1 * p[i]
            else:
                solution.objectives[i] = p[i]
        return solution

    def evaluator(self, candidates, *args):
        res = []
        for candidate in candidates:
            res.append(self.evaluate(candidate))
        return res

    def get_name(self) -> str:
        return self.problem.get_name()

    def build_operators(self):
        from .operators import build_ko_operators
        return build_ko_operators(self.problem)


class JMetalOUProblem(Problem[OUSolution], Evaluable):

    def __init__(self, problem, initial_polulation=[]):
        """JMetal OU problem. Encapsulates a MEWpy problem
        so that it can be used in jMetal.
        """
        self.problem = problem
        self.number_of_objectives = len(self.problem.fevaluation)
        self.obj_directions = []
        self.obj_labels = []
        for f in self.problem.fevaluation:
            self.obj_labels.append(str(f))
            if f.maximize:
                self.obj_directions.append(self.MAXIMIZE)
            else:
                self.obj_directions.append(self.MINIMIZE)
        self.initial_polulation = initial_polulation
        self.__next_ini_sol = 0

    def create_solution(self) -> OUSolution:
        solution = None
        flag = False
        while self.__next_ini_sol < len(self.initial_polulation) and not flag:
            s = self.initial_polulation[self.__next_ini_sol]
            try:
                solution = self.problem.encode(s)
                flag = True
                self.__next_ini_sol += 1
            except ValueError as e:
                print("Skipping seed:", s, " ", e)
                self.__next_ini_sol += 1
        if not solution:
            solution = self.problem.generator(random)
        new_solution = OUSolution(
            self.problem.bounder.lower_bound,
            self.problem.bounder.upper_bound,
            len(solution),
            self.problem.number_of_objectives)
        new_solution.variables = list(solution)[:]
        return new_solution

    def reset_initial_population_counter(self):
        import random
        random.shuffle(self.initial_polulation)
        self.__next_ini_sol = 0

    def get_constraints(self, solution):
        return self.problem.decode(set(solution.variables))

    def evaluate(self, solution: KOSolution) -> KOSolution:
        candidate = set(solution.variables)
        p = self.problem.evaluate_solution(candidate)
        for i in range(len(p)):
            # JMetalPy only deals with minimization problems
            if self.obj_directions[i] == self.MAXIMIZE:
                solution.objectives[i] = -1 * p[i]
            else:
                solution.objectives[i] = p[i]
        return solution

    def evaluator(self, candidates, *args):
        res = []
        for candidate in candidates:
            res.append(self.evaluate(candidate))
        return res

    def get_name(self) -> str:
        return self.problem.get_name()

    def build_operators(self):
        from .operators import build_ou_operators
        return build_ou_operators(self.problem)



from .gecko import GeckoKOProblem, GeckoOUProblem
from .genes import GKOProblem, GOUProblem
from .reactions import RKOProblem, ROUProblem, MediumProblem
from .etfl import ETFLGKOProblem, ETFLGOUProblem
from .optram import load_optram, OptRAMRegModel, OptRamProblem
from .optorf import load_optorf, OptORFProblem
from .com import CommunityKOProblem
from .kinetic import KineticKOProblem, KineticOUProblem



"""JMetalpy operators
"""

import copy
import random
from typing import List

from jmetal.core.operator import Mutation, Crossover
from jmetal.core.solution import Solution

from .problem import KOSolution, OUSolution
from ...util.constants import EAConstants


class ShrinkMutation(Mutation[Solution]):
    """ Shrink mutation. A gene is removed from the solution.

    :param probability: (float), The mutation probability.
    :param min_size: (int) the solution minimum size.

    """

    def __init__(self, probability: float = 0.1, min_size: int = EAConstants.MIN_SOLUTION_SIZE):
        super(ShrinkMutation, self).__init__(probability=probability)
        self.min_size = min_size

    def execute(self, solution: Solution) -> Solution:
        """
        Apply the mutation.

        :param solution: The candidate solution to be mutated.
        :returns: A mutated solution.

        """
        if random.random() <= self.probability and solution.number_of_variables > self.min_size:
            var = copy.copy(solution.variables)
            index = random.randint(0, len(var) - 1)
            del var[index]
            solution.variables = var
            solution.number_of_variables = len(var)
        return solution

    def get_name(self):
        return 'Shrink Mutation'


class GrowMutationKO(Mutation[KOSolution]):
    """ Grow mutation. A gene is added to the solution.

    :param probability: (float), The mutation probability.
    :param min_size: (int) the solution minimum size.

    """

    def __init__(self, probability: float = 0.1, max_size: int = EAConstants.MAX_SOLUTION_SIZE):
        super(GrowMutationKO, self).__init__(probability=probability)
        self.max_size = max_size

    def execute(self, solution: Solution) -> Solution:
        """
        Apply the mutation.

        :param solution: The candidate solution to be mutated.
        :returns: A mutated solution.

        """
        if random.random() <= self.probability and solution.number_of_variables < self.max_size:
            mutant = copy.copy(solution.variables)
            newElem = random.randint(solution.lower_bound, solution.upper_bound)
            while newElem in mutant:
                newElem = random.randint(solution.lower_bound, solution.upper_bound)
            mutant.append(newElem)
            solution.variables = mutant
            solution.number_of_variables = len(mutant)
        return solution

    def get_name(self):
        return 'Grow Mutation KO'


class GrowMutationOU(Mutation[OUSolution]):
    """ Grow mutation. A gene is added to the solution.

    :param probability: (float), The mutation probability.
    :param min_size: (int) the solution minimum size.

    """

    def __init__(self, probability: float = 0.1, max_size: int = EAConstants.MAX_SOLUTION_SIZE):
        super(GrowMutationOU, self).__init__(probability=probability)
        self.max_size = max_size

    def execute(self, solution: Solution) -> Solution:
        """
        Apply the mutation.

        :param solution: The candidate solution to be mutated.
        :returns: A mutated solution.

        """
        if random.random() <= self.probability and solution.number_of_variables < self.max_size:
            mutant = copy.copy(solution.variables)
            idx = random.randint(solution.lower_bound[0], solution.upper_bound[0])
            idxs = [a for (a, b) in mutant]
            while idx in idxs:
                idx = random.randint(solution.lower_bound[0], solution.upper_bound[0])
            lv = random.randint(solution.lower_bound[1], solution.upper_bound[1])
            mutant.append((idx, lv))
            solution.variables = mutant
            solution.number_of_variables = len(mutant)
        return solution

    def get_name(self):
        return 'Grow Mutation OU'


class UniformCrossoverKO(Crossover[KOSolution, KOSolution]):
    """Uniform Crossover for KO solutions

    :param probability: (float) The probability of crossover.
    :param max_size: (int) The solution maximum size.

    """

    def __init__(self, probability: float = 0.1, max_size: int = EAConstants.MAX_SOLUTION_SIZE):
        super(UniformCrossoverKO, self).__init__(probability=probability)
        self.max_size = max_size

    def execute(self, parents: List[KOSolution]) -> List[KOSolution]:
        if len(parents) != 2:
            raise Exception('The number of parents is not two: {}'.format(len(parents)))

        offspring = [copy.deepcopy(parents[0]), copy.deepcopy(parents[1])]

        if random.random() <= self.probability and (
                offspring[0].number_of_variables > 1 or offspring[1].number_of_variables > 1):
            mom = set(copy.copy(offspring[0].variables))
            dad = set(copy.copy(offspring[1].variables))
            intersection = mom & dad
            otherElems = list((mom | dad).difference(intersection))
            child1 = copy.copy(intersection)
            child2 = copy.copy(intersection)

            while len(otherElems) > 0:
                elemPosition = random.randint(0, len(otherElems) - 1) if len(otherElems) > 1 else 0
                if len(child1) == self.max_size or len(child2) == 0:
                    child2.add(otherElems[elemPosition])
                elif len(child2) == self.max_size or len(child1) == 0:
                    child1.add(otherElems[elemPosition])
                else:
                    r = random.random()
                    if r <= 0.5:
                        child1.add(otherElems[elemPosition])
                    else:
                        child2.add(otherElems[elemPosition])
                otherElems.pop(elemPosition)

            offspring[0].variables = list(child1)
            offspring[0].number_of_variables = len(child1)
            offspring[1].variables = list(child2)
            offspring[1].number_of_variables = len(child2)
        return offspring

    def get_number_of_parents(self) -> int:
        return 2

    def get_number_of_children(self) -> int:
        return 2

    def get_name(self):
        return 'Uniform Crossover KO'


class MutationContainer(Mutation[Solution]):
    """A container for the mutation operators.

    :param probability: (float) The probability of applying a mutation.
    :param mutators: (list) The list of mutators.

    """

    def __init__(self, probability: float = 0.5, mutators=[]):
        super(MutationContainer, self).__init__(probability=probability)
        self.mutators = mutators

    def execute(self, solution: Solution) -> Solution:
        # randomly select a mutator and apply it
        if random.random() <= self.probability:
            idx = random.randint(0, len(self.mutators) - 1)
            mutator = self.mutators[idx]
            return mutator.execute(solution)
        else:
            return solution

    def get_name(self):
        return 'Mutation container'


class UniformCrossoverOU(Crossover[OUSolution, OUSolution]):
    """
        Uniform Crossover for OU solutions
    """

    def __init__(self, probability: float = 0.1, max_size: int = EAConstants.MAX_SOLUTION_SIZE):
        super(UniformCrossoverOU, self).__init__(probability=probability)
        self.max_size = max_size

    def execute(self, parents: List[KOSolution]) -> List[KOSolution]:
        if len(parents) != 2:
            raise Exception('The number of parents is not two: {}'.format(len(parents)))

        offspring = [copy.deepcopy(parents[0]), copy.deepcopy(parents[1])]

        if random.random() <= self.probability and (
                offspring[0].number_of_variables > 1 or offspring[1].number_of_variables > 1):
            mom = set(copy.copy(offspring[0].variables))
            dad = set(copy.copy(offspring[1].variables))
            intersection = mom & dad
            otherElems = list((mom | dad).difference(intersection))
            child1 = copy.copy(intersection)
            child2 = copy.copy(intersection)

            while len(otherElems) > 0:
                elemPosition = random.randint(0, len(otherElems) - 1) if len(otherElems) > 1 else 0
                if len(child1) == self.max_size or len(child2) == 0:
                    child2.add(otherElems[elemPosition])
                elif len(child2) == self.max_size or len(child1) == 0:
                    child1.add(otherElems[elemPosition])
                else:
                    r = random.random()
                    if r <= 0.5:
                        child1.add(otherElems[elemPosition])
                    else:
                        child2.add(otherElems[elemPosition])
                otherElems.pop(elemPosition)

            offspring[0].variables = list(child1)
            offspring[0].number_of_variables = len(child1)
            offspring[1].variables = list(child2)
            offspring[1].number_of_variables = len(child2)
        return offspring

    def get_number_of_parents(self) -> int:
        return 2

    def get_number_of_children(self) -> int:
        return 2

    def get_name(self):
        return 'Uniform Crossover OU'


class SingleMutationKO(Mutation[KOSolution]):
    """
    Mutates a single element
    """

    def __init__(self, probability: float = 0.1):
        super(SingleMutationKO, self).__init__(probability=probability)

    def execute(self, solution: Solution) -> Solution:
        if random.random() <= self.probability:
            mutant = copy.copy(solution.variables)
            index = random.randint(0, len(mutant) - 1)
            newElem = random.randint(solution.lower_bound, solution.upper_bound)
            while newElem in mutant:
                newElem = random.randint(solution.lower_bound, solution.upper_bound)
            mutant[index] = newElem
            solution.variables = mutant
        return solution

    def get_name(self):
        return 'Single Mutation KO'


class SingleMutationOU(Mutation[OUSolution]):
    """
    Mutates a single element
    """

    def __init__(self, probability: float = 0.1):
        super(SingleMutationOU, self).__init__(probability=probability)

    def execute(self, solution: Solution) -> Solution:
        if random.random() <= self.probability:
            mutant = copy.copy(solution.variables)
            lix = [i for (i, j) in mutant]
            index = random.randint(0, len(mutant) - 1)
            idx, idy = mutant[index]
            is_mutate_idx = False
            if random.random() > 0.5:
                idx = random.randint(solution.lower_bound[0], solution.upper_bound[0])
                while idx in lix:
                    idx = random.randint(solution.lower_bound[0], solution.upper_bound[0])
                is_mutate_idx = True
            lv = random.randint(solution.lower_bound[1], solution.upper_bound[1])
            if not is_mutate_idx:
                while lv == idy:
                    lv = random.randint(solution.lower_bound[1], solution.upper_bound[1])
            mutant[index] = (idx, lv)
            solution.variables = mutant
        return solution

    def get_name(self):
        return 'Single Mutation KO'


class SingleMutationOULevel(Mutation[OUSolution]):
    """
    Mutates the expression level of a single element
    """

    def __init__(self, probability: float = 0.1):
        super(SingleMutationOULevel, self).__init__(probability=probability)

    def execute(self, solution: Solution) -> Solution:
        if random.random() <= self.probability:
            mutant = copy.copy(solution.variables)
            index = random.randint(0, len(mutant) - 1)
            idx, idy = mutant[index]
            lv = random.randint(solution.lower_bound[1], solution.upper_bound[1])
            while lv == idy:
                lv = random.randint(solution.lower_bound[1], solution.upper_bound[1])
            mutant[index] = (idx, lv)
            solution.variables = mutant
        return solution

    def get_name(self):
        return 'Single Mutation KO'


class GaussianMutation(Mutation[Solution]):
    """
     A Gaussian mutator
    """

    def __init__(self, probability: float = 0.1,
                 gaussian_mutation_rate: float = 0.1,
                 gaussian_mean: float = 0.0,
                 gaussian_std: float = 1.0):
        super(GaussianMutation, self).__init__(probability=probability)
        self.gaussian_mutation_rate = gaussian_mutation_rate
        self.gaussian_mean = gaussian_mean
        self.gaussian_std = gaussian_std

    def execute(self, solution: Solution) -> Solution:
        if random.random() <= self.probability:
            mutant = copy.copy(solution.variables)
            for i, m in enumerate(mutant):
                if random.random() < self.gaussian_mutation_rate:
                    v = m + random.gauss(self.gaussian_mean, self.gaussian_std)
                    while v < solution.lower_bound[i] or v > solution.upper_bound[i]:
                        v = m + random.gauss(self.gaussian_mean, self.gaussian_std)
                    solution.variables[i] = v
        return solution

    def get_name(self):
        return 'Gaussian Mutator'

class UniformCrossover(Crossover[Solution, Solution]):
    """
        Uniform Crossover
    """

    def __init__(self, probability: float = 0.1):
        super(UniformCrossover, self).__init__(probability=probability)

    def execute(self, parents: List[Solution]) -> List[Solution]:
        if len(parents) != 2:
            raise Exception('The number of parents is not two: {}'.format(len(parents)))

        offspring = [copy.deepcopy(parents[0]), copy.deepcopy(parents[1])]

        if random.random() <= self.probability and (
                offspring[0].number_of_variables > 1 or offspring[1].number_of_variables > 1):
            mom = copy.copy(offspring[0].variables)
            dad = copy.copy(offspring[1].variables)
            child1 = []
            child2 = []
            for p in range(len(mom)):
                if random.random() <= 0.5:
                    child1.append(mom[p])
                    child2.append(dad[p])
                else:
                    child1.append(dad[p])
                    child2.append(mom[p])
                
            offspring[0].variables = list(child1)
            offspring[0].number_of_variables = len(child1)
            offspring[1].variables = list(child2)
            offspring[1].number_of_variables = len(child2)
        return offspring

    def get_number_of_parents(self) -> int:
        return 2

    def get_number_of_children(self) -> int:
        return 2

    def get_name(self):
        return 'Uniform Crossover'



class SingleRealMutation(Mutation[Solution]):
    """
    Mutates a single element
    """

    def __init__(self, probability: float = 0.1):
        super(SingleRealMutation, self).__init__(probability=probability)

    def execute(self, solution: Solution) -> Solution:
        if random.random() <= self.probability:
            index = random.randint(0, solution.number_of_variables - 1)
            solution.variables[index] = solution.lower_bound[index] + \
                (solution.upper_bound[index] - solution.lower_bound[index]) * random.random()
        return solution

    def get_name(self):
        return 'Single Real Mutation'


REP_INT = {
    "SHRINK": ShrinkMutation,
    "GROWKO": GrowMutationKO,
    "GROWOU": GrowMutationOU,
    "UCROSSKO": UniformCrossoverKO,
    "UCROSSOU": UniformCrossoverOU,
    "SMUTKO": SingleMutationKO,
    "SMUTOU": SingleMutationOU,
    "SMLEVEL": SingleMutationOULevel
}


def build_ko_operators(problem):
    crossover = UniformCrossoverKO(0.8, problem.candidate_max_size)
    mutators = []

    # add shrink and growth mutation only if max size != min size
    if problem.candidate_max_size != problem.candidate_min_size:
        mutators.append(GrowMutationKO(
            1.0, max_size=problem.candidate_max_size))
        mutators.append(ShrinkMutation(
            1.0, min_size=problem.candidate_min_size))

    mutators.append(SingleMutationKO(1.0))
    mutations = MutationContainer(0.3, mutators=mutators)
    return crossover, mutations


def build_ou_operators(problem):
    crossover = UniformCrossoverOU(0.5, problem.candidate_max_size)
    mutators = []
    _max = problem.candidate_max_size
    _min = problem.candidate_min_size
    _t_size = problem.bounder.upper_bound[0]+1

    # add shrink and growth mutation only if max size != min size
    # and do not add if single ou if  max size == min size == targets size
    if _max != _min:
        mutators.append(GrowMutationOU(
            1.0, max_size=problem.candidate_max_size))
        mutators.append(ShrinkMutation(
            1.0, min_size=problem.candidate_min_size))
        mutators.append(SingleMutationOU(1.0))
    elif _min != _t_size:
        mutators.append(SingleMutationOU(1.0))
    else:
        pass

    mutators.append(SingleMutationOULevel(1.0))
    mutations = MutationContainer(0.3, mutators=mutators)
    return crossover, mutations
